'use strict';

var launchApp = angular.module("launchApp", [
	"commonApp"
]);

launchApp.directive('launchScrollDetect',function(){
    return {
	    restrict: "A",
	    link: function(scope, element, attributes) {
	        if (scope.$last) {
	        	scope.$emit("RefreshScroll");
	        }
	    }
	}
});

launchApp.directive('launchAutoScroll', function() {
	return {
	    restrict: "A",
	    link: function(scope, element, attributes) {
	        scope.$on("RefreshScroll",function() {
        		element[0].scrollTop = element[0].scrollHeight;
	        });
	    }
	}
});

launchApp.directive('launchDetails', function() {
	return {
	    restrict: "A",
	    scope: {
	    	launch: '=launch',
	    	maxHeight: '=launchMaxHeight'
	    },
	    templateUrl: "../common/templates/launch-details.html"
	}
});

launchApp.factory("LaunchLogEntry", 
	["Resource", 
	function(Resource) {
		
	return {
		create: function() {
			var resource = Resource.create();
			
			resource.getMessage = function() {
				return resource.data.message;
			};
			
			resource.getDuration = function() {
				return MDW.utils.Date.toDurationString(resource.data.duration);
			};
			
			resource.getClass = function() {
				return "launch-log-" + resource.data.kind;
			};
			
			return resource;
		}
	}
}]);

launchApp.factory("Launch", 
	["LaunchLogEntry", "Resource", 
	function(LaunchLogEntry, Resource) {
		
	return {
		create: function() {
			var resource = Resource.create();
			var progress = null;
			var logs = [];
			
			resource.addListener({
				loadData: function(data) {
					
					// last progress message
					var getLaunchProgressInfo = function() {
						
				    	// waiting?
				    	if (data.state == "waiting") {
				    		return "Preparing execution...";
				    	}
				    	
				    	// running -> use last progress info
				    	if (data.state == "running") {
					    	if (data.logEntries) {
					    		for (var i = data.logEntries.length - 1; i >= 0; i--) {
					    			if (data.logEntries[i].kind == "progress") {
					    				return data.logEntries[i].message;
					    			}
					    		}
					    	}
					    	
							return "Executing...";
				    	}
				    	
				    	else {
				    		return null;
				    	}
				    };
				    progress = getLaunchProgressInfo();
				    
				    // logs
				    logs = resource.createResources(data.logEntries, LaunchLogEntry);
				}
			});

			resource.getId = function() { return this.data.id; };
			resource.getUrl = function() { return this.data.uri; };
			resource.getServiceId = function() { return this.data.service; };
			resource.getProgress = function() { return progress; };
			resource.getLogs = function() { return logs; };
			
			resource.isTerminated = function() { 
				return resource.data.state != "waiting" && resource.data.state != "running";
			};
			
			resource.isCanceled = function() { 
				return resource.data.state == "canceled";
			};
			
			return resource;
		}
	}
}]);
