'use strict';

var stateApp = angular.module("stateApp", [
     "commonApp"
]);

stateApp.controller("StateCtrl", 
	["$scope", "$http", "$timeout", "ServerState", 
	function($scope, $http, $timeout, ServerState) {
		
	var refreshPromise = null;
	$scope.initialized = false;
	$scope.state = ServerState.create();
	$scope.resumeOnRestart = $scope.state.isResumeOnRestart();
	
	$scope.state.addListener({
		loadData: function() {
			$scope.initialized = true;
			
			// synchronize the resumeOnRestart value 
			$scope.resumeOnRestart = $scope.state.isResumeOnRestart();
			
			// auto-refresh as long as maintenance is requested and the queue is not empty
			if (!$scope.state.isActive() && $scope.state.getRunningLaunchCount() > 0) {
				refreshPromise = $timeout($scope.actions.refresh.run, 3000);
			}
		}
	});
	
	var requestStateChange = function(value) {
		$http({
			method: "PUT",
			url: $scope.state.getUrl(),
			params: {
				value: value,
				resumeOnRestart: $scope.resumeOnRestart
			}
		}).then(function(response) {
			$scope.state.loadResource({ url: $scope.state.getUrl() });
		});
	};

	$scope.actions = {
		refresh: {
			isDisabled: function() {
				return $scope.state.operation.isRunning();
			},
			run: function() {
				if ($scope.actions.refresh.isDisabled()) {
					return;
				}
				if (refreshPromise) {
					$timeout.cancel(refreshPromise);
					refreshPromise = null;
				}
				$scope.state.loadResource({url: "../../server/state"});
			}
		},
		requestActive: {
			isDisabled: function() {
				return $scope.state.operation.isRunning();
			},
			run: function() {
				if ($scope.actions.requestActive.isDisabled()) {
					return;
				}
				requestStateChange("active");
			}
		},
		requestSuspended: {
			isDisabled: function() {
				return $scope.state.operation.isRunning();
			},
			run: function() {
				if ($scope.actions.requestSuspended.isDisabled()) {
					return;
				}
				requestStateChange("suspended");
			}
		},
		requestMaintenance: {
			isDisabled: function() {
				return $scope.state.operation.isRunning();
			},
			run: function() {
				if ($scope.actions.requestMaintenance.isDisabled()) {
					return;
				}
				requestStateChange("maintenance");
			}
		},
		toggleResumeOnRestart: {
			isDisabled: function() {
				return $scope.state.operation.isRunning();
			},
			run: function() {
				if ($scope.actions.toggleResumeOnRestart.isDisabled()) {
					return;
				}
				$scope.resumeOnRestart = !$scope.resumeOnRestart;
				requestStateChange("maintenance");
			}
		}
	};
		
	$scope.actions.refresh.run();
}]);
