'use strict';

var servicesApp = angular.module("servicesApp", [
 	"commonApp",
 	"launchApp"
]);

function getPropertyValue(parameter, name) {
	if (parameter.properties && parameter.properties.length != 0) {
    	for (var i = 0; i < parameter.properties.length; i++) {
    		if (name == parameter.properties[i].name) {
    			return parameter.properties[i].value;
    		}
    	}
	}
	
	return null;
};

function onStorageTypeChange(parameterName) {
	var select = $("#" + parameterName + "_select");  
	
	// remove current input
	select.siblings("input").remove();
	
	// add new input
	if (select.val() == "file") {
		select.parent().append(getTextInputFieldHtml(parameterName, "file"));
	}
	else {
		select.parent().append(getTextInputFieldHtml(parameterName, "text"));
	}
}

function getTextInputFieldHtml(name, fieldType) {
	return "<input type='" + fieldType + "' name='" + name + "' id='input_'" + name + "' class='form-control input-sm'/>";
}

function getBooleanInputFieldHtml(name) {
	return "<input class='radioMargin' type='checkbox' name='" + name + "' id='input_'" + name + "' value='true'/>" + name;
}

function getStorageInputFieldHtml(name, isIn) {
	var buffer = "";
	var inputHml = null;
	buffer += "<div class='storage'><select id='" + name + "_select' class='form-control input-sm' onchange='onStorageTypeChange(\"" + name + "\")'>";
	
	// in direction -> file upload
	if (isIn) {
		buffer += "<option selected value='file'>File</option>";
		inputHml = getTextInputFieldHtml(name, "file");
		buffer += "<option value='uri'>URI</option>";
	}
	else {
		buffer += "<option selected value='uri'>URI</option>";
		inputHml = getTextInputFieldHtml(name, "text");
	}
	
	buffer += "</select>";
	buffer += inputHml;
	buffer += "</div>"
	
	return buffer;
}

function getInputFieldHtml(parameter) {
	
	// storage?
	if (parameter.type == "com.sodius.mdw.server.storage.Storage") {
		
		// out?
		if ("out" == getPropertyValue(parameter, "direction")) {
			return getStorageInputFieldHtml(parameter.name, false /* out */);
		}
		else {
			return getStorageInputFieldHtml(parameter.name, true /* in */);
		}
	}
	
	// boolean?
	if (parameter.type == "boolean") {
		return getBooleanInputFieldHtml(parameter.name);
	}
	
	// look for "password" property
	if ("true" == getPropertyValue(parameter, "password")) {
		return getTextInputFieldHtml(parameter.name, "password");
	}
	
	// "password" name (for compatibility only)
	if (parameter.name.toLowerCase().indexOf("password") >= 0) {
		return getTextInputFieldHtml(parameter.name, "password");
	}
	
	return getTextInputFieldHtml(parameter.name, "text");
}

servicesApp.controller("ServicesCtrl", 
	["$scope", "$http", 
	function($scope, $http) {
		
	$scope.loading = true;
	$scope.error = null;
	$scope.services = [];
	$scope.selectedServices = [];
	$scope.launchLocation = null;
	
	$scope.state = 0; // 0 = service selection, 1 = service arguments
	
	// collect services
	$http.get("../../service")
		.then(function(res) {
			$scope.loading = false;
			$scope.services = res.data.services;                
		}, function(error) {
			$scope.loading = false;
			$scope.error = error.data;
		});
	
	// listen on service selection change to update arguments
	$scope.$watch('selectedServices', function(newValue, oldValue) {
		$scope.onServiceChange();
	});
	
	$scope.isStateServices = function() {
		return $scope.state == 0;
	}
	
	$scope.isStateArguments = function() {
		return $scope.state == 1;
	}
	
	$scope.showServices = function() {
		$scope.state = 0;
	};
	
	$scope.showArguments = function() {
		$scope.state = 1;
	};
	
	$scope.getService = function() {
		if ($scope.selectedServices.length == 1) {
			return $scope.selectedServices[0];
		}
		else {
			return null;
		}
	};

	/*
	 * Rebuild the launch argument list
	 */
	$scope.onServiceChange = function() {
    	var nodeParameters = $("#parameters");
		var service = $scope.getService();
		if (service == null) {
			nodeParameters.empty();
			return;
		}

		// register launch action on form submit
		$('#formLaunch').ajaxForm({
			url: service.launchFactory,
			success: function(data, textStatus, jqXHR) {
				refreshState($scope.getContentLocation(jqXHR));
			},
			error: function(jqXHR, textStatus, errorThrown) {
				if (jqXHR.status != 200) {
					alert(jqXHR.responseText);
				}
			}
		}); 
    	
		// update parameters
    	nodeParameters.empty();
		if (service.parameters) {
	    	$.each(service.parameters, function() {
	    		var buffer = "";
	    		buffer += "<div class='form-group'>";
	    		buffer += "<label class='control-label'>" + this.name + " <small>(" + this.type + ")</small><\/label>";
	    		buffer += "<div>";
	    		buffer += getInputFieldHtml(this);
	    		buffer += "<\/div>";
	    		buffer += "<\/div>";
	    		nodeParameters.append(buffer);
	    	});
		}
	};
}]);


servicesApp.controller("LaunchCtrl", 
	["$scope", "$http", "Launch",
	function($scope, $http, Launch) {
		
	$scope.title = null;
	$scope.launch = Launch.create();
	
	$('#executionDialog').on('shown.bs.modal', function () {
		$scope.title = "Executing " + $scope.getService().name;
		$scope.$apply();
		
		// valid?
		if ($scope.errorMessage == null) {
			$scope.progressMessage = "Preparing execution...";
			$scope.execute();
		}
		
		else {
			$scope.title = "Execution Aborted";
			$scope.$apply();
		}
	});
	
	$('#executionDialog').on('hidden.bs.modal', function () {
		$scope.launch.loadData({});
		$scope.errorMessage = null;
		$scope.progressMessage = null;
		$scope.launchLocation = null;
		$scope.cancelRequested = false;
		$scope.$apply();
	});
	
	var getContentLocation = function(xhr) {
		
		// Content-Location in header
		var location = xhr.getResponseHeader("Content-Location");
		
		// Content-Location in response body (the header might not be seen using older browsers)
		if (! location) {
			location = xhr.responseText;
			
			// there might be surrounding <pre> </pre>
			if (location && location.length > 11 && location.slice(0, 5).toLowerCase() == "<pre>") {
				location = location.slice(5, location.length - 6);
			}
		}
		
		return location;
	};

	$scope.execute = function() {
		
		$('#formLaunch').ajaxSubmit({
			url: $scope.getService().launchFactory,
			success: function(data, statusText, xhr) {
				$scope.progressMessage = null;
				$scope.launchLocation = getContentLocation(xhr);
				$scope.$apply();
				$scope.refreshState();
			},
			error: function (xhr, ajaxOptions, thrownError) {
				$scope.progressMessage = null;
				$scope.errorMessage = thrownError || "Request failed";
				$scope.$apply();
		    }
		}); 
	}
	
	$scope.refreshState = function() {
		if (! $scope.launchLocation) {
			return;
		}
		
		$scope.launch.loadResource({
			url: $scope.launchLocation,
			success: function(response) {
				if ($scope.launch.isTerminated()) {
					if ($scope.launch.isCanceled()) {
                		$scope.title = "Execution Cancelled";
    				}
    				else {
                		$scope.title = "Execution Terminated";
    				}
				}
				else {
					// refresh in 800 milliseconds
	                setTimeout(function() {
	                	$scope.refreshState();
	                }, 800);
				}
			}
		});
	}
	
	$scope.cancelExecution = function() {
		if ($scope.launch.isTerminated()) {
			$('#executionDialog').modal("hide");
		}
		else if ($scope.launchLocation && ! $scope.cancelRequested) {
			$scope.cancelRequested = true;
			$scope.launch.deleteResource({});
		}
	}
}]);