pragma runLim, 0 
pragma encoding, "UTF-8"

bool ignoreError = false;
Module openedModule = null; 
Skip moduleNamesToClose = createString();
Skip targetModules = createString();
Skip linkModulesFullNames = createString();
Skip createdObjectsSkip = create();
Skip absnoObjectsSkip = create();
Skip createdDiscussionsSkip = create();
Object luo = null;
Buffer outputBuffer = create();

void sodius_finalize() {
	if (! null moduleNamesToClose) {
		delete(moduleNamesToClose);
		moduleNamesToClose = null;
	}
	
	if (! null createdObjectsSkip) {
		delete(createdObjectsSkip);
		createdObjectsSkip = null;
	}
	
	if (! null absnoObjectsSkip) {
		delete(absnoObjectsSkip);
		absnoObjectsSkip = null;
	}
	
	if (! null createdDiscussionsSkip) {
		delete(createdDiscussionsSkip);
		createdDiscussionsSkip = null;
	}
	
	if (! null outputBuffer) {
		delete(outputBuffer);
		outputBuffer = null;
	}
	
	if (! null targetModules) {
		delete(targetModules);
		targetModules = null;
	}
	
	if (! null linkModulesFullNames) {
		delete(linkModulesFullNames);
		linkModulesFullNames = null;
	}
}

void sodius_writeOutput() {
	
	// verify the output is not already written
	if (null outputFilename) {
		return;
	}
	
	Stream file = write(outputFilename, CP_UTF8);
	file << outputBuffer;
	close(file);
	
	outputFilename = null;
}

void sodius_setOutput(int requestId, string message) {
	int l = length(message);
	outputBuffer += "" requestId "|" l "|" message "|";
}

void sodius_closeModules() {
	if (null moduleNamesToClose) {
		return;
	}
	
	string qualifiedName;
	for qualifiedName in moduleNamesToClose do {
		Item i = item(qualifiedName);
		if (! null i) {
			Module m = module(i);
			if (! null m) {
				
				// link module that wasn't open at first and which is now edit -> save
				if (type(i) == "Link" && isEdit(m)) {
					save(m);
				}
				
				close(m, false);
			}
		}
	}
}

void sodius_discardAndClose() {
	if (! null openedModule) {
		close(openedModule, false);
		openedModule = null;
	}
	
	sodius_closeModules();
	sodius_writeOutput();
	sodius_finalize();
}

void sodius_setResult(string message) {
	
	// verify a message is not already logged
	if (null resultFilename) {
		return;
	}
	
	Stream file = write(resultFilename, CP_UTF8);
	file << message;
	close(file);
	
	resultFilename = null;
}

bool sodius_isTerminated() {
	return null resultFilename;
}

void sodius_setErrorIgnored(bool ignored) {
	ignoreError = ignored;
}

void sodius_setError(string message) {
	if (ignoreError) {
		return;
	}
	
	sodius_discardAndClose();
	
	if (null message || length(message) == 0) {
		sodius_setResult("Empty DOORS error message");
	}
	else {
		sodius_setResult(message);
	}
}

string sodius_getMessageOrLastError(string message) {
	string l = lastError();
	if (null message) {
		return message;
	}
	else {
		return l;
	}
}

void sodius_deleteItem(Item i) {
	if (null i) {
		return;
	}
	
	// close the open module that might correspond to this item
	Module m = module(i);
	if (! null m) {
		close(m, false);
		m = null;
	}
	
	openedModule = null;
	
	string message = delete(i);
	if (! null message) {
		sodius_setError("Cannot delete item '" fullName(i) "': " message);
		return;
	}
}

void sodius_purgeItem(Item i) {
	if (null i) {
		return;
	}
	
	string message = purge(i);
	if (! null message) {
		sodius_setError("Cannot purge item '" fullName(i) "': " message);
		return;
	}
}

void sodius_leaveModuleOpen() {
	openedModule = null;
	sodius_closeModules();
	sodius_writeOutput();
	sodius_finalize();
}

void sodius_saveModule_(bool closeOnSave) {
	if (! null openedModule) {
		flushDeletions();
		save(openedModule);
		if (closeOnSave) {
			close(openedModule);
		}
		openedModule = null;
	}
	
	sodius_closeModules();
	sodius_writeOutput();
	sodius_finalize();
}

void sodius_saveModuleAndLeaveOpen() {
	sodius_saveModule_(false);
}

void sodius_saveAndCloseModule() {
	sodius_saveModule_(true);
}

void sodius_archiveModule(Item i, string archiveFilename) {
	if (null i) {
		return;
	}
	
	string message = archive(module fullName(i), archiveFilename, false);
	
	if (! null message) {
		sodius_setError("Cannot archive module '" fullName(i) "': " message);
		return;
	}
	
	if (! canOpenFile(archiveFilename, false)) {
		sodius_setError("Cannot archive module '" fullName(i) "': unknown reason");
		return;
	}
}

void sodius_archiveProject(Item i, string archiveFilename) {
	if (null i) {
		return;
	}
	
	string message = archive(fullName(i), archiveFilename, false);
	
	if (! null message) {
		sodius_setError("Cannot archive project '" fullName(i) "': " message);
		return;
	}
	
	if (! canOpenFile(archiveFilename, false)) {
		sodius_setError("Cannot archive project '" fullName(i) "': unknown reason");
		return;
	}
}

void sodius_restoreModule(Item i, string archiveFilename, string moduleName) {
	if (null i) {
		return;
	}
	
	Folder f = folder(i);
	current = f;
	
	string message;
	if (null moduleName) {
		message = restoreModule(archiveFilename);
	}
	else {
		message = restoreModule(archiveFilename, moduleName);
	}

	if (! null message) {
		sodius_setError("Cannot restore module archive into '" fullName(i) "': " message);
		return;
	}
}

void sodius_restoreProject(Item i, string archiveFilename, string projectName, string projectDescription) {
	if (null i) {
		return;
	}
	
	Folder f = folder(i);
	current = f;
	
	string message;
	if (null projectName) {
		message = restoreProject(archiveFilename);
	}
	else {
		message = restoreProject(archiveFilename, projectName, projectDescription);
	}

	if (! null message) {
		sodius_setError("Cannot restore project archive into '" fullName(i) "': " message);
		return;
	}
}

bool sodius_isDoorsClientCompliantWith(char majorVersion, char minorVersion) {
	char actualMajorVersion = '0';
	char actualMinorVersion = '0';
	string sVersion = doorsInfo(infoVersion) 
	if (length(sVersion) > 2) {
	    actualMajorVersion = sVersion[0];
	    actualMinorVersion = sVersion[2];
	}
	else {
		// weird, can't tell the version number
		return false;
	}
	
	if (actualMajorVersion < majorVersion) {
		return false;
	}
	else if (actualMajorVersion == majorVersion) {
		return actualMinorVersion >= minorVersion;
	}
	else {
		return true;
	}
}

// If DOORS client is >= 9.4, there is a DXL function to decode a Resource URL.
string decodeResourceURL_URL = null; 
bool decodeResourceURL_decoded = false; 
Item decodeResourceURL_Item = null; 
int decodeResourceURL_absno = 0; 
string sodius_decodeResourceURLSource() {
	if (sodius_isDoorsClientCompliantWith('9', '4')) {
		return "
		string protocol;
		string dbHost;
		int dbPort;
		string repositoryId;
		string dbName;
		string dbId;
		ModuleVersion modVer;
		string viewName;
		
    	string &decodeResourceURL_URL = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_URL)) "); 
    	bool &decodeResourceURL_decoded = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_decoded)) "); 
    	Item &decodeResourceURL_Item = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_Item)) ");
    	int &decodeResourceURL_absno = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_absno)) ");
	
		string message = null;
		noError;
		message = decodeResourceURL(decodeResourceURL_URL, protocol, dbHost, dbPort, repositoryId, dbName, dbId, 
			decodeResourceURL_Item, modVer, viewName, decodeResourceURL_absno);
		string lastErrorMessage = lastError; 
		decodeResourceURL_decoded = null message && null lastErrorMessage;
		";
	}
	else {
		return "";
	}
}
string decodeResourceURL_sourceCode = sodius_decodeResourceURLSource();

bool sodius_decodeResourceURL(string url, Item &resultItem, int &absno) {
	if (length(decodeResourceURL_sourceCode) == 0) {
		return false;
	}
	else {
		decodeResourceURL_URL = url;
		decodeResourceURL_decoded = false;
		decodeResourceURL_absno = 0;
		
		eval_ decodeResourceURL_sourceCode;
		
		bool result = decodeResourceURL_decoded;
		if (result) {
			resultItem = decodeResourceURL_Item;
			absno = decodeResourceURL_absno;
		}
		
		decodeResourceURL_URL = null;
		decodeResourceURL_decoded = false;
		decodeResourceURL_absno = 0;
		
		return result;
	}
}

bool sodius_decodeURL(string url, Item &resultItem, int &absno) {
	string dbHost;
	int dbPort;
	string dbName;
	string dbId;
	ModuleVersion modVer;
	
	noError;
	string message = decodeURL(url, dbHost, dbPort, dbName, dbId, resultItem, modVer, absno);
	string lastErrorMessage = lastError;
	bool result = null message && null lastErrorMessage;
	return result;
}

Item sodius_getItemByURL(string url) {
	Item resultItem;
	int absno;
	if (sodius_decodeURL(url, resultItem, absno)) {
		return resultItem;
	}
	else if (sodius_decodeResourceURL(url, resultItem, absno)) {
		return resultItem;
	}
	else {
		sodius_setError("The DOORS Item does not exist: " url); //DoorsStatusCode
		return null;
	}
}

Item sodius_getItemByUniqueID(string itemId) {
	Item i = itemFromID(itemId);
	if (null i) {
		sodius_setError("The DOORS Item does not exist: " itemId); //DoorsStatusCode
		return null;
	}
	else {
		return i;
	}
}

Item sodius_getItemByQualifiedName(string qualifiedName) {
	Item i = item(qualifiedName);
	if (null i) {
		sodius_setError("The DOORS Item does not exist: " qualifiedName); //DoorsStatusCode
		return null;
	}
	else {
		return i;
	}
}

Item sodius_getItemByName(Item parentItem, string name) {
	if (null parentItem) {
		return null;
	}
	else {
		return sodius_getItemByQualifiedName(fullName(parentItem) "/" name);
	}
}

string sodius_getParentPath(string qualifiedName) {
	int i = 0;
	for (i = length(qualifiedName); i > 0; i--) {
		if (qualifiedName[i] == '/') {
			return qualifiedName[0:i - 1];
		}
	}
	return "";
}

void sodius_createFolderByQualifiedName(string qualifiedName) {
	
	// item already exists?
	Item i = item(qualifiedName);
	if (! null i) {
		if (type(i) == "Folder" || type(i) == "Project") {
			return;
		}
		else {
			sodius_setError("The Item '" qualifiedName "' is an existing module");
			return;
		}
	}
	
	// create the folder
	Folder f;
	string message = create(qualifiedName, "", f);
    if (! null message) {
    	sodius_setError("Cannot create folder '" qualifiedName "': " message);
		return;
    }
    else if (null f) {
    	sodius_setError("Cannot create folder '" qualifiedName "': unknown reason, might be related to an invalid folder name");
		return;
    }
}

void sodius_createFolderByName(Item parentItem, string name) {
	if (null parentItem) {
		return;
	}
	else {
		sodius_createFolderByQualifiedName(fullName(parentItem) "/" name);
	}
}

void sodius_createProjectByQualifiedName(string qualifiedName) {
	
	// item already exists?
	Item i = item(qualifiedName);
	if (! null i) {
		if (type(i) == "Folder" || type(i) == "Project") {
			return;
		}
		else {
			sodius_setError("The Item '" qualifiedName "' is an existing module");
			return;
		}
	}
	
	// create the project
	Project p;
	string message = create(qualifiedName, "", p);
    if (! null message) {
    	sodius_setError("Cannot create project '" qualifiedName "': " message);
		return;
    }
    else if (null p) {
    	sodius_setError("Cannot create project '" qualifiedName "': unknown reason, might be related to an invalid project name");
		return;
    }
}

void sodius_createProjectByName(Item parentItem, string name) {
	if (null parentItem) {
		return;
	}
	else {
		sodius_createProjectByQualifiedName(fullName(parentItem) "/" name);
	}
}

string sodius_getDefaultLinkModuleQualifiedName(Item sourceModule, Item targetModule) {
	if (null sourceModule || null targetModule) {
		return "";
	}
	string linkModuleName = getDefaultLinkModule(module(fullName(sourceModule)), module(fullName(targetModule)));
	if (null linkModuleName || length(linkModuleName) == 0) {
		linkModuleName = "DOORS Links";
	}
	string parentPath = sodius_getParentPath(linkModuleName);
	if (null parentPath || length(parentPath) == 0) {
		parentPath = sodius_getParentPath(fullName(sourceModule));
		linkModuleName = parentPath "/" linkModuleName;
	}
	return linkModuleName;
}

void sodius_createModuleByQualifiedName(string qualifiedName) {
	
	// check the parent folder exists
	string parentPath = sodius_getParentPath(qualifiedName);
	if (null parentPath || length(parentPath) == 0) {
		sodius_setError("A qualified name must be used to create a module: " qualifiedName);
		return;
	}
	Folder f = folder(parentPath);
	if (null f) {
		sodius_setError("The folder does not exist: " parentPath);
		return;
	}
	
	// create module
	noError;
	openedModule = create(qualifiedName, "", "", 1, false);
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot create module '" qualifiedName "': " message);
		return;
    }
	else if (null openedModule) {
		sodius_setError("Cannot create the module: " qualifiedName);
		return;
	}
}

void sodius_createModuleByName(Item parentItem, string name) {
	if (null parentItem) {
		return;
	}
	else {
		sodius_createModuleByQualifiedName(fullName(parentItem) "/" name);
	}
}

void sodius_registerLinkModule(string variableName, string qualifiedName) {
	put(linkModulesFullNames, variableName, qualifiedName);
	
	Item i = item(qualifiedName)
	if (! null i) {
		Module m = module(i);
		if (! null m) {
			put(moduleNamesToClose, qualifiedName, qualifiedName);
		}
	}
}

void sodius_createLinkModuleByQualifiedName(string qualifiedName, int linkKind, string variableName) {
	
	// just verify the module does exist?
	if (! null variableName) {
		Item i = item(qualifiedName);
		if (! null i && type(i) == "Link") {
			sodius_registerLinkModule(variableName,qualifiedName);
			return;
		}
	}
	
	// check the parent folder exists
	string parentPath = sodius_getParentPath(qualifiedName);
	if (null parentPath || length(parentPath) == 0) {
		sodius_setError("A qualified name must be used to create a module: " qualifiedName);
		return;
	}
	Folder f = folder(parentPath);
	if (null f) {
		sodius_setError("The folder does not exist: " parentPath);
		return;
	}
	
	// create module
	noError;
	Module m = create(qualifiedName, "", linkKind, false);
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot create module '" qualifiedName "': " message);
		return;
    }
	else if (null m) {
		sodius_setError("Cannot create the module: " qualifiedName);
		return;
	}
	else {
		if (null variableName) {
			openedModule = m;
		}
		else {
			sodius_registerLinkModule(variableName, qualifiedName);
		}
	}
}

void sodius_createLinkModuleByName(Item parentItem, string name, int linkKind, string variableName) {
	if (null parentItem) {
		return;
	}
	else {
		sodius_createLinkModuleByQualifiedName(fullName(parentItem) "/" name, linkKind, variableName);
	}
}

void sodius_finalizeOpenModule() {
	// loop on all objects to get their absolute number
	Object o;
	for o in entire openedModule do {
	    if (! isDeleted(o)) {
    		int absoluteNumber = o."Absolute Number";
    		bool result = put(absnoObjectsSkip, absoluteNumber, o);
    		if (! result) {
    			sodius_setError("Cannot register a object from its absolute number: " absoluteNumber "");
    			return;
    		}
	    }
	}
}

void sodius_readModule(Item i) {
    if (null i) {
        return;
    }
    
    openedModule = module(i);
    if (null openedModule) {
        noError;
        openedModule = read(fullName(i), false /* no display */, true /* standard view */);
        lastError; // the module may have DXL errors we are not interested in
    }
    else {
        refresh(openedModule);
    }
    
    if (null openedModule) {
        sodius_setError("Cannot open the module in read-only mode: " fullName(i)); //[DoorsStatusFactory]
    }
    else {
        sodius_finalizeOpenModule();
    }
}

void sodius_editModule(Item i) {
    if (null i) {
        return;
    }
    
    // if the module is currently open in read mode, close it so that it can then be open in edit mode (MMDOORS-87)
    openedModule = module(i);
    if (! null openedModule && isRead(openedModule)) {
        close(openedModule, false);
        openedModule = null;
    }
    
    noError;
    openedModule = edit(fullName(i), false /* no display */, true /* standard view */);
    lastError; // the module may have DXL errors we are not interested in
    if (null openedModule) {
        sodius_setError("Cannot open the module in exclusive edit mode: " fullName(i)); //[DoorsStatusFactory]
    }
    else {
        sodius_finalizeOpenModule();
    }
}

void sodius_shareModule(Item i) {
	if (null i) {
		return;
	}
	
	noError;
	openedModule = share(fullName(i), false /* no display */, true /* standard view */);
	lastError; // the module may have DXL errors we are not interested in
	if (null openedModule) {
		sodius_setError("Cannot open the module in share mode: " fullName(i)); //[DoorsStatusFactory]
	}
	else {
		sodius_finalizeOpenModule();
	}
}

void sodius_getOpenModule(Item i) {
	if (null i) {
		return;
	}
	
	openedModule = module(i);
}

void sodius_requestSwitchToEdit(int requestorAbsno, string message, bool canSwitchToEdit) {
	if (canSwitchToEdit) {
		if (unsaved(openedModule)) {
			if (requestorAbsno == 0) {
				sodius_setError("Cannot open the module with unsaved changes in edit mode: " fullName(openedModule)); //[DoorsStatusFactory]
			}
			else {
				sodius_setError("Cannot lock object " requestorAbsno " and cannot open the module with unsaved changes in edit mode: " message); //[DoorsStatusFactory]
			}
		}
		else {
			string fullname = fullName(openedModule);
			close(openedModule, false);
			absnoObjectsSkip = create();
			noError;
			openedModule = edit(fullname, false /* no display */, true /* standard view */);
			lastError; // the module may have DXL errors we are not interested in
			if (null openedModule) {
				if (requestorAbsno == 0) {
					sodius_setError("Cannot open the module in exclusive edit mode: " fullname); //[DoorsStatusFactory]
				}
				else {
					sodius_setError("Cannot lock object " requestorAbsno " and cannot open the module in edit mode: " message); //[DoorsStatusFactory]
				}
			}
			else {
				sodius_finalizeOpenModule();
			}
		}
	}
	else {
		if (requestorAbsno == 0) {
			sodius_setError("Cannot modify the module top content in share mode");
		}
		else {
			sodius_setError("Cannot lock object " requestorAbsno ": " message); //[DoorsStatusFactory]
		}
	}
}

void sodius_lockObject(Object self, bool switchToEdit) {
	if (null self) {
		return;
	}
	if (isLockedByUser(self)) {
		return;
	}
	string message = lock(self);
	if (! null message) {
		int absno = self."Absolute Number";
		sodius_requestSwitchToEdit(absno, message, switchToEdit);
	}
}

void sodius_lockParentObject(Object self, bool switchToEdit) {
	if (null self) {
		return;
	}
	Object parentObject = parent(self);
	if (null parentObject) {
		sodius_requestSwitchToEdit(0, null, switchToEdit);
	}
	else {
		sodius_lockObject(parentObject, switchToEdit);
	}
}

void sodius_output_getAttributeValue(int requestId, Object o, string attributeName) {
	if (null o) {
		return;
	}
	
	noError(); 
    string value = "" o.attributeName "";
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot get attribute '" attributeName "' of object " o."Absolute Number" ": " message);
    	return;
    }
    
    sodius_setOutput(requestId, value); 
}

void sodius_output_getIdentifierValue(int requestId, Object o) {
	if (null o) {
		return;
	}
    
    sodius_setOutput(requestId, identifier(o)); 
}

void sodius_output_getURLValue(int requestId, Object o) {
	if (null o) {
		return;
	}
    
    sodius_setOutput(requestId, sodius_getURL(o)); 
}

void sodius_output_getModuleAttributeValue(int requestId, string attributeName) {
	noError(); 
    string value = "" openedModule.attributeName "";
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot get module attribute '" attributeName "': " message);
    	return;
    }
    
    sodius_setOutput(requestId, value); 
}

void sodius_output_getModuleURLValue(int requestId) {
    sodius_setOutput(requestId, sodius_getModuleURL(openedModule)); 
}

void sodius_output_getLinkAttributeValue(int requestId, Link l, string attributeName) {
	if (null l) {
		return;
	}
	
	noError(); 
    string value = "" l.attributeName "";
    
    string message = lastError();
    if (! null message) {
    	ModName_ modname = target(l);
    	sodius_setError("Cannot get attribute '" attributeName "' of link between Object " sourceAbsNo(l) " and Object " targetAbsNo(l) " (module: " fullName(modname) "): " message);
    	return;
    }
    
    sodius_setOutput(requestId, value); 
}

void sodius_createType(string name, AttrBaseType baseType) {
	current = openedModule;
	noError;
	
	string message = null;
	create(name, baseType, message);
	message = sodius_getMessageOrLastError(message);
	if (! null message) {
    	sodius_setError("Cannot create type '" name "': " message);
    	return;
	}
}

AttrType sodius_getType(string typeName) {
	AttrType t = find(openedModule, typeName);
	if (null t) {
		sodius_setError("The DOORS type does not exist: " typeName);
		return null;
	}
	else {
		return t;
	}
}

void sodius_createTypeInt(string name, int min, int max) {
	current = openedModule;
	noError;
	
	string message = null;
	create(name, min, max, message);
	message = sodius_getMessageOrLastError(message);
	if (! null message) {
    	sodius_setError("Cannot create type '" name "': " message);
    	return;
	}
}

void sodius_createTypeReal(string name, real min, real max) {
	current = openedModule;
	noError;
	
	string message = null;
	create(name, min, max, message);
	message = sodius_getMessageOrLastError(message);
	if (! null message) {
    	sodius_setError("Cannot create type '" name "': " message);
    	return;
	}
}

void sodius_createTypeDate(string name, Date min, Date max) {
	current = openedModule;
	noError;
	
	string message = null;
	create(name, min, max, message);
	message = sodius_getMessageOrLastError(message);
	if (! null message) {
    	sodius_setError("Cannot create type '" name "': " message);
    	return;
	}
}

void sodius_createTypeEnumeration(string name, string codes[]) {
	current = openedModule;
	noError;
	
	string message = null;
	create(name, codes, message);
	message = sodius_getMessageOrLastError(message);
	if (! null message) {
    	sodius_setError("Cannot create type '" name "': " message);
    	return;
	}
}

void sodius_modifyType(AttrType t, string p_name, AttrBaseType p_baseType) {
	if (null t) {
		return;
	}
	
	if (! null p_name) {
		noError;
		string message = null;
		modify(t, p_name, message);
		message = sodius_getMessageOrLastError(message);
		if (! null message) {
	    	sodius_setError("Cannot modify name of type '" t.name "': " message);
	    	return;
		}
	}

	if (p_baseType == attrDate || p_baseType == attrInteger || p_baseType == attrReal || p_baseType == attrText || p_baseType == attrString || p_baseType == attrUsername || p_baseType == attrEnumeration) {
		noError;
		string message = null;
		modify(t, p_baseType, message);
		message = sodius_getMessageOrLastError(message);
		if (! null message) {
	    	sodius_setError("Cannot modify base type of type '" t.name "': " message);
	    	return;
		}
	}
}
	
void sodius_modifyTypeLiterals(AttrType t, string codes[]) {
	int i = 0;
	int s = sizeof(codes);
	int values[s];
	int colors[s];
	for (i = 0; i < s; i++) {
		values[i] = i;
		colors[i] = -1;
	}
	
	noError;
	string message = null;
	modify(t, t.name, codes, values, colors, message);
	message = sodius_getMessageOrLastError(message);
	if (! null message) {
    	sodius_setError("Cannot modify enumeration literals of type '" t.name "': " message);
    	return;
	}
}

void sodius_modifyTypeIntMin(AttrType t, int p_min) {
	if (null t) {
		return;
	}
	
	noError;
	setMinValue(t, p_min, true);
	string message = lastError();
	if (! null message) {
    	sodius_setError("Cannot modify minimum value of type '" t.name "': " message);
    	return;
	}
}

void sodius_modifyTypeIntMax(AttrType t, int p_max) {
	if (null t) {
		return;
	}
	
	noError;
	setMaxValue(t, p_max, true);
	string message = lastError();
	if (! null message) {
    	sodius_setError("Cannot modify maximum value of type '" t.name "': " message);
    	return;
	}
}

void sodius_modifyTypeRealMin(AttrType t, real p_min) {
	if (null t) {
		return;
	}
	
	noError;
	setMinValue(t, p_min, true);
	string message = lastError();
	if (! null message) {
    	sodius_setError("Cannot modify minimum value of type '" t.name "': " message);
    	return;
	}
}

void sodius_modifyTypeRealMax(AttrType t, real p_max) {
	if (null t) {
		return;
	}
	
	noError;
	setMaxValue(t, p_max, true);
	string message = lastError();
	if (! null message) {
    	sodius_setError("Cannot modify maximum value of type '" t.name "': " message);
    	return;
	}
}

void sodius_modifyTypeDateMin(AttrType t, Date p_min) {
	if (null t) {
		return;
	}
	
	noError;
	setMinValue(t, p_min, true);
	string message = lastError();
	if (! null message) {
    	sodius_setError("Cannot modify minimum value of type '" t.name "': " message);
    	return;
	}
}

void sodius_modifyTypeDateMax(AttrType t, Date p_max) {
	if (null t) {
		return;
	}
	
	noError;
	setMaxValue(t, p_max, true);
	string message = lastError();
	if (! null message) {
    	sodius_setError("Cannot modify maximum value of type '" t.name "': " message);
    	return;
	}
}

void sodius_deleteType(string typeName) {
	AttrType t = sodius_getType(typeName);
	if (null t) {
		return;
	}
	
	string message = null;
    if (! delete(t, message)) {
    	if (null message) {
    		sodius_setError("Cannot delete type '" typeName "': unknown reason");
    	}
    	else {
    		sodius_setError("Cannot delete type '" typeName "': " message);
    	}
    	return;
    }
}

AttrDef sodius_getAttribute(string attributeName) {
	AttrDef a = find(openedModule, attributeName);
	if (null a) {
		sodius_setError("The DOORS Attribute does not exist: " attributeName);
		return null;
	}
	else {
		return a;
	}
}

AttrDef sodius_createAttribute(string attributeName, string typeName, bool forObject, bool forModule) {
	current = openedModule;
	noError;
	
	AttrDef result = null;
	if (forObject && forModule) {
		result = create module object type typeName attribute attributeName
	}
	else if (forObject) {
		result = create object type typeName attribute attributeName
	}
	else {
		result = create module type typeName attribute attributeName
	}
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot create attribute '" attributeName "' of type '" typeName "': " message);
    	return null;
    }
    return result;
}

void sodius_modifyAttribute(AttrDef self, string p_default, string p_dxl, string p_name, int p_bars, int p_date, int p_history, int p_inherit, int p_forModule, int p_multi, int p_forObjects) {
	if (null self) {
		return;
	}
	
	if (! null p_default) { self = modify(self, setDefault, p_default); }
	if (! null p_dxl) { self = modify(self, setDXL, p_dxl); }
	if (! null p_name && self.name != p_name) { self = modify(self, setName, p_name); }
	if (p_bars != -1) { self = modify(self, setBars, p_bars == 1); }
	if (p_date != -1) { self = modify(self, setDates, p_date == 1); }
	if (p_history != -1) { self = modify(self, setHistory, p_history == 1); }
	if (p_inherit != -1) { self = modify(self, setInherit, p_inherit == 1); }
	if (p_forModule != -1) { self = modify(self, setModule, p_forModule == 1); }
	if (p_multi != -1) { self = modify(self, setMulti, p_multi == 1); }
	if (p_forObjects != -1) { self = modify(self, setObject, p_forObjects == 1); }
}

void sodius_deleteAttribute(string attributeName) {
	AttrDef a = sodius_getAttribute(attributeName);
	if (null a) {
		return;
	}
	
	string message = delete(openedModule, a);
    if (! null message) {
    	sodius_setError("Cannot delete attribute '" attributeName "': " message);
    	return;
    }
}

Object sodius_getCreatedObject(int index) {
	if (sodius_isTerminated()) {
		return null;
	}
	
	Object o = null;
	bool result = find(createdObjectsSkip, index, o);
	if (result) {
		luo = o;
		sodius_output_getAttributeValue(index, o, "Absolute Number");
		return o;
	}
	else {
		sodius_setError("Cannot find a created object: " index "");
		return null;
	}
}

Object sodius_getParentObject(Object o) {
    if (null o) {
        return null;
    }
    else {
        Object p = parent(o);
        if (null p) {
            sodius_setError("The object has no parent: " o."Absolute Number" "");
            return null;
        }
        else {
            luo = p;
            return p;
        }
    }
}

Object sodius_findObject(int absoluteNumber) {
	if (sodius_isTerminated()) {
		return null;
	}
	
	Object o = null;
	bool result = find(absnoObjectsSkip, absoluteNumber, o);
	if (result) {
		luo = o;
		return o;
	}
	else {
		if (type(item(fullName(openedModule))) == "Link") {
			sodius_setError("Invalid command, a link module does not contain objects");
			return null;
		}
		else {
			sodius_setError("The absolute number does not match an existing object: " absoluteNumber ""); //DoorsStatusCode
			return null;
		}
	}
}

Object sodius_findObjectByURL(string url) {
	if (sodius_isTerminated()) {
		return null;
	}
	
	Item resultItem;
	int absno;
	if (sodius_decodeURL(url, resultItem, absno)) {
	    if (absno > 0) {
	        return sodius_findObject(absno);
	    }
	}
	else if (sodius_decodeResourceURL(url, resultItem, absno)) {
	    if (absno > 0) {
	        return sodius_findObject(absno);
	    }
	}
	
	sodius_setError("The URL does not match an existing object: " url ""); //DoorsStatusCode
	return null;
}

Object sodius_verifyMoveTargetObject(Object targetObject) {
	if (null targetObject) {
		return null;
	}
    else if (cell(targetObject)) {
		sodius_setError("Cannot move an object relatively to a cell: " targetObject."Absolute Number" "");
		return null;
    }
    else if (row(targetObject)) {
		sodius_setError("Cannot move an object relatively to a row: " targetObject."Absolute Number" "");
		return null;
    }
    else {
    	return targetObject;
    }
}

Object sodius_verifyMoveObject(Object objectToMove) {
	if (null objectToMove) {
		return null;
	}
    else if (cell(objectToMove)) {
		sodius_setError("Cannot move a cell: " objectToMove."Absolute Number" "");
		return null;
    }
    else if (row(objectToMove)) {
		sodius_setError("Cannot move a row: " objectToMove."Absolute Number" "");
		return null;
    }
    else {
    	return objectToMove;
    }
}

void sodius_moveObjectBefore(Object objectToMove, Object targetObject) {
	objectToMove = sodius_verifyMoveObject(objectToMove);
	targetObject = sodius_verifyMoveTargetObject(targetObject);
	if (null objectToMove || null targetObject || objectToMove == targetObject)  {
		return;
	}
	
	// Need to move object after the first object, and then reverse the positions
	// (Can't find a 'move before' function)
	noError
	move(objectToMove, targetObject);
	move(targetObject, objectToMove);
	string message = lastError
	if (! null message) {
		sodius_setError("Cannot move object " objectToMove."Absolute Number" " before object " targetObject."Absolute Number" ": " message);
		return;
	}
}

void sodius_moveObjectAfter(Object objectToMove, Object targetObject) {
	objectToMove = sodius_verifyMoveObject(objectToMove);
	targetObject = sodius_verifyMoveTargetObject(targetObject);
	if (null objectToMove || null targetObject || objectToMove == targetObject)  {
		return;
	}
	
	noError
	move(objectToMove, targetObject);
	string message = lastError
	if (! null message) {
		sodius_setError("Cannot move object " objectToMove."Absolute Number" " after object " targetObject."Absolute Number" ": " message);
		return;
	}
}

void sodius_moveObjectTop(Object objectToMove) {
    objectToMove = sodius_verifyMoveObject(objectToMove);
    Object targetObject = first(openedModule);
    sodius_moveObjectBefore(objectToMove, targetObject);
}

void sodius_moveObjectLastTop(Object objectToMove) {
    Object predecessor = null;
    for predecessor in top(openedModule) do {}
    sodius_moveObjectAfter(objectToMove, predecessor);
}

void sodius_moveObjectBelow(Object objectToMove, Object targetObject) {
    objectToMove = sodius_verifyMoveObject(objectToMove);
    targetObject = sodius_verifyMoveTargetObject(targetObject);
    if (null objectToMove || null targetObject || objectToMove == targetObject)  {
        return;
    }
    
    noError
    move(objectToMove, below(targetObject));
    string message = lastError
    if (! null message) {
        sodius_setError("Cannot move object " objectToMove."Absolute Number" " below object " targetObject."Absolute Number" ": " message);
        return;
    }
}

void sodius_moveObjectBelowParent(Object objectToMove, Object targetChild) {
    if (null objectToMove || null targetChild)  {
        return;
    }
    Object targetObject = parent(targetChild);
    if (null targetObject) {
        sodius_moveObjectTop(objectToMove);
    }
    else {
        sodius_moveObjectBelow(objectToMove, targetObject);
    }
}

void sodius_moveObjectLastBelow(Object objectToMove, Object targetObject) {
    objectToMove = sodius_verifyMoveObject(objectToMove);
    targetObject = sodius_verifyMoveTargetObject(targetObject);
    if (null objectToMove || null targetObject || objectToMove == targetObject)  {
        return;
    }
    
    noError
    move(objectToMove, last(below(targetObject)));
    string message = lastError
    if (! null message) {
        sodius_setError("Cannot move object " objectToMove."Absolute Number" " below object " targetObject."Absolute Number" ": " message);
        return;
    }
}

void sodius_moveObjectLastBelowParent(Object objectToMove, Object targetChild) {
    if (null objectToMove || null targetChild)  {
        return;
    }
    Object targetObject = parent(targetChild);
    if (null targetObject) {
        sodius_moveObjectLastTop(objectToMove);
    }
    else {
        sodius_moveObjectLastBelow(objectToMove, targetObject);
    }
}

Object sodius_registerCreatedObject(Object o, int index) {
	bool result = put(createdObjectsSkip, index, o);
	if (result) {
		sodius_output_getAttributeValue(index, o, "Absolute Number");
		return o;
	}
	else {
		sodius_setError("Cannot register a created object: " index "");
		return null;
	}
}

Object sodius_registerPotentiallyCreatedObject(Object o, int index, string location, Object siblingObject, string message) {
    if (null o) {
        Buffer b = create();
        b += "Cannot create an object " location;
        if (null siblingObject) {
            b += " " fullName(openedModule);
        }
        else {
            b += " object " siblingObject."Absolute Number" ""
        }
        if (! null message) {
            b += ": " message;
        }
        sodius_setError(tempStringOf(b));
        delete(b);
        return null;
    }
    else {
        return sodius_registerCreatedObject(o, index);
    }
}

Object sodius_createObjectTop(int index) {
    noError
    Object o = create(openedModule);
    string message = lastError
    return sodius_registerPotentiallyCreatedObject(o, index, "top in module", null, message);
}

Object sodius_createObjectBefore(int index, Object siblingObject) {
	if (null siblingObject) {
		return null;
	}
	
	noError
	Object o = create(before(siblingObject));
	string message = lastError
	return sodius_registerPotentiallyCreatedObject(o, index, "before", siblingObject, message);
}

Object sodius_createObjectAfter(int index, Object siblingObject) {
	if (null siblingObject) {
		return null;
	}
	
	noError
	Object o = create(after(siblingObject));
	string message = lastError
    return sodius_registerPotentiallyCreatedObject(o, index, "after", siblingObject, message);
}

Object sodius_createObjectLastTop(int index) {
    Object predecessor = null;
    for predecessor in top(openedModule) do {}
    if (null predecessor) {
        return sodius_createObjectTop(index);
    }
    else {
        return sodius_createObjectAfter(index, predecessor);
    }
}

Object sodius_createObjectBelow(int index, Object siblingObject) {
	if (null siblingObject) {
		return null;
	}

	noError
	Object o = create(below(siblingObject));
	string message = lastError
    return sodius_registerPotentiallyCreatedObject(o, index, "below", siblingObject, message);
}

Object sodius_createObjectLastBelow(int index, Object siblingObject) {
    if (null siblingObject) {
        return null;
    }

    noError
    Object o = create(last(below(siblingObject)));
    string message = lastError
    return sodius_registerPotentiallyCreatedObject(o, index, "below", siblingObject, message);
}

Object sodius_createPictureTop(int index, string filepath, int pictureFormat) {
	// Creating a picture as first object in the module fails with an exception access violation.
	// The workaround is to create an dummy object, insert the picture after it, and then remove this dummy object
	
	Object siblingObject = create(openedModule);
	if (null siblingObject) {
		sodius_setError("Cannot create a top object in module: " fullName(openedModule));
		return null;
	}
	else {
		if (insertPictureFileAfter(filepath, pictureFormat, siblingObject)) {
			Object result = sodius_registerCreatedObject(next(sibling(siblingObject)), index);
			hardDelete(siblingObject);
			return result;
		}
		else {
			hardDelete(siblingObject);
			sodius_setError("Cannot insert picture as a top object in the module");
			return null;
		}
	}
}

Object sodius_createPictureAfter(int index, string filepath, int pictureFormat, Object siblingObject) {
	if (null siblingObject) {
		return sodius_createPictureTop(index, filepath, pictureFormat);
	}

	if (insertPictureFileAfter(filepath, pictureFormat, siblingObject)) {
		return sodius_registerCreatedObject(next(sibling(siblingObject)), index);
	}
	else {
		sodius_setError("Cannot insert picture after: " siblingObject."Absolute Number" "");
		return null;
	}
}

Object sodius_createPictureBefore(int index, string filepath, int pictureFormat, Object siblingObject) {
	Object o = sodius_createPictureAfter(index, filepath, pictureFormat, siblingObject);
	if (null o) {
		return null;
	}
	else {
		sodius_moveObjectBefore(o, siblingObject);
		return o;
	}
}

Object sodius_createPictureLastTop(int index, string filepath, int pictureFormat) {
    Object predecessor = null;
    for predecessor in top(openedModule) do {}
    if (null predecessor) {
        return sodius_createPictureTop(index, filepath, pictureFormat);
    }
    else {
        return sodius_createPictureAfter(index, filepath, pictureFormat, predecessor);
    }
}

Object sodius_createPictureBelow(int index, string filepath, int pictureFormat, Object siblingObject) {
    if (null siblingObject) {
        return null;
    }

    if (insertPictureFileBelow(filepath, pictureFormat, siblingObject)) {
        return sodius_registerCreatedObject(first(siblingObject), index);
    }
    else {
        sodius_setError("Cannot insert picture below: " siblingObject."Absolute Number" "");
        return null;
    }
}

Object sodius_createPictureLastBelow(int index, string filepath, int pictureFormat, Object siblingObject) {
    Object o = sodius_createPictureBelow(index, filepath, pictureFormat, siblingObject);
    if (null o) {
        return null;
    }
    else {
        sodius_moveObjectLastBelow(o, siblingObject);
        return o;
    }
}

void sodius_replacePicture(Object o, string filepath, int pictureFormat) {
	if (null o) {
		return;
	}
	
	int oldPictureFormat = getPictFormat(o);
	if (oldPictureFormat > 0) {
		if (insertPictureFile(filepath, pictureFormat, o)) {
			return;
		}
		else {
			sodius_setError("Cannot replace picture for object: " o."Absolute Number" "");
		}
	}
	else {
		sodius_setError("Cannot replace picture of a non-picture object: " o."Absolute Number" "");
	}
}

void sodius_deletePicture(Object o) {
	if (null o) {
		return;
	}
	
	if (getPictFormat(o) != 0) {
		if (! deletePicture(o)) {
			sodius_setError("Cannot delete picture of a object: " o."Absolute Number" "");
		}
	}
}

Object sodius_verifyCreateTable(Object targetObject) {
	if (null targetObject) {
		return null;
	}
    else if (cell(targetObject)) {
		sodius_setError("Cannot create a table relatively to a cell: " targetObject."Absolute Number" "");
		return null;
    }
    else if (row(targetObject)) {
		sodius_setError("Cannot create a table relatively relatively to a row: " targetObject."Absolute Number" "");
		return null;
    }
    else {
    	return targetObject;
    }
}

void sodius_formatTable(Object tableObject) {
	if (null tableObject) {
		return;
	}
	
	setAllCellsBorder(tableObject, left, solidBorder);
	setAllCellsBorder(tableObject, right, solidBorder);
	setAllCellsBorder(tableObject, top, solidBorder);
	setAllCellsBorder(tableObject, bottom, solidBorder);
}

Object sodius_registerPotentiallyCreatedTable(Object o, int index, string location, Object siblingObject, string message) {
    if (null o) {
        Buffer b = create();
        b += "Cannot create a table " location;
        if (null siblingObject) {
            b += " " fullName(openedModule);
        }
        else {
            b += " object " siblingObject."Absolute Number" ""
        }
        if (! null message) {
            b += ": " message;
        }
        sodius_setError(tempStringOf(b));
        delete(b);
        return null;
    }
    else {
        o = sodius_registerCreatedObject(o, index);
        sodius_formatTable(o);
        return o;
    }
}

Object sodius_createTableTop(int index, int rows, int cols) {
	noError
	Object tableObject = table(openedModule, rows, cols);
	string message = lastError
	return sodius_registerPotentiallyCreatedTable(tableObject, index, "top in module", null, message);
}

Object sodius_createTableBefore(int index, int rows, int cols, Object parentObject) {
	parentObject = sodius_verifyCreateTable(parentObject);
	if (null parentObject) {
		return;
	}
	
	noError
	Object tableObject = table(before(parentObject), rows, cols);
	string message = lastError
    return sodius_registerPotentiallyCreatedTable(tableObject, index, "before", parentObject, message);
}

Object sodius_createTableAfter(int index, int rows, int cols, Object parentObject) {
	parentObject = sodius_verifyCreateTable(parentObject);
	if (null parentObject) {
		return;
	}
	
	noError
	Object tableObject = table(after(parentObject), rows, cols);
	string message = lastError
    return sodius_registerPotentiallyCreatedTable(tableObject, index, "after", parentObject, message);
}

Object sodius_createTableBelow(int index, int rows, int cols, Object parentObject) {
    parentObject = sodius_verifyCreateTable(parentObject);
    if (null parentObject) {
        return;
    }
    
    noError
    Object tableObject = table(below(parentObject), rows, cols);
    string message = lastError
    return sodius_registerPotentiallyCreatedTable(tableObject, index, "below", parentObject, message);
}

Object sodius_createTableLastBelow(int index, int rows, int cols, Object parentObject) {
    parentObject = sodius_verifyCreateTable(parentObject);
    if (null parentObject) {
        return;
    }
    
    noError
    Object tableObject = table(last(below(parentObject)), rows, cols);
    string message = lastError
    return sodius_registerPotentiallyCreatedTable(tableObject, index, "below", parentObject, message);
}

Object sodius_createTableLastTop(int index, int rows, int cols) {
    Object predecessor = null;
    for predecessor in top(openedModule) do {}
    if (null predecessor) {
        return sodius_createTableTop(index, rows, cols);
    }
    else {
        return sodius_createTableAfter(index, rows, cols, predecessor);
    }
}

Object sodius_getTableFromObject(Object o) {
	if (null o) {
		return null;
	}
	else if (table(o)) {
		return o;
	}
	else if (cell(o)) {
		return sodius_getTableFromObject(getTable(o));
	}
	else if (row(o)) {
		return sodius_getTableFromObject(parent(o));
	}
	else {
		sodius_setError("This object is not contained in a table: " o."Absolute Number" "");
		return null;
	}
}

int sodius_getRowCount(Object tableObject) {
	int rows = 0;
	Object rowObject;
	for rowObject in table(tableObject) do {
		if (! isDeleted(rowObject)) {
			rows++;
		}
	}
	return rows;
}

int sodius_getColumnCount(Object tableObject) {
	Object rowObject;
	Object cellObject
	for rowObject in table(tableObject) do {
		if (! isDeleted(rowObject)) {
			int columns = 0;
			for cellObject in row(rowObject) do {
				if (! isDeleted(cellObject)) {
					columns++;
				}
			}
			return columns;
		}
	}
	return 0;
}

Object sodius_getLastTableCell(Object tableObject) {
	Object rowObject;
	Object cellObject;
	Object lastCell = null;
	for rowObject in table(tableObject) do {
		if (! isDeleted(rowObject)) {
			for cellObject in row(rowObject) do {
				if (! isDeleted(cellObject)) {
					lastCell = cellObject;
				}
			}
		}
	}
	return lastCell;
}

void sodius_resizeTable(Object tableObject, int rows, int columns) {
	tableObject = sodius_getTableFromObject(tableObject);
	if (null tableObject) {
		return;
	}
	int i = 0;
	
	// rows
	int rowCount = sodius_getRowCount(tableObject);
	if (rowCount < rows) {
		for (i = rows - rowCount; i > 0; i--) {
			noError;
			appendRow(sodius_getLastTableCell(tableObject));
		    string message = lastError();
		    if (! null message) {
		    	sodius_setError("Cannot append row to table " tableObject."Absolute Number" ": " message);
		    }
		}
	}
	else if (rowCount > rows) {
		for (i = rowCount - rows ; i > 0; i--) {
			string message = deleteRow(sodius_getLastTableCell(tableObject));
			if (! null message) {
				sodius_setError("Failed to resize table " tableObject."Absolute Number" ": " message);
				return;
			}
		}
	}

	// columns
	int columnCount = sodius_getColumnCount(tableObject);
	if (columnCount < columns) {
		for (i = columns - columnCount; i > 0; i--) {
			noError;
			appendColumn(sodius_getLastTableCell(tableObject));
		    string message = lastError();
		    if (! null message) {
		    	sodius_setError("Cannot append column to table " tableObject."Absolute Number" ": " message);
		    }
		}
	}
	else if (columnCount > columns) {
		for (i = columnCount - columns; i > 0; i--) {
			string message = deleteColumn(sodius_getLastTableCell(tableObject));
			if (! null message) {
				sodius_setError("Failed to resize table " tableObject."Absolute Number" ": " message);
				return;
			}
		}
	}
}

Object sodius_getTableRow(Object tableObject, int rowIndex) {
	if (null tableObject) {
		return null;
	}
	else {
		Object rowObject;
		int index = 0;
		for rowObject in table(tableObject) do {
			if (rowIndex == index) {
				return rowObject;
			}
			else {
				index++;
			}
		}
		sodius_setError("No row at index " rowIndex " for table: " tableObject."Absolute Number" "");
		return null;
	}
}

Object sodius_getRowCell(Object rowObject, int columnIndex) {
	if (null rowObject) {
		return null;
	}
	else {
		Object cellObject;
		int index = 0;
		for cellObject in row(rowObject) do {
			if (columnIndex == index) {
				return cellObject;
			}
			else {
				index++;
			}
		}
		sodius_setError("No cell at index " columnIndex " for row: " rowObject."Absolute Number" "");
		return null;
	}
}

Object sodius_getTableCell(Object tableObject, int rowIndex, int columnIndex) {
	Object o = sodius_getRowCell(sodius_getTableRow(tableObject, rowIndex), columnIndex);
	if (null o) {
		return null;
	}
	else {
		luo = o;
		return o;
	}
}

void sodius_setColumnWidths(Object tableObject, int columnWidths[]) {
	if (null tableObject) {
		return;
	}
	
	Object rowObject = null;
	for rowObject in table(tableObject) do {
		
		int i = 0;
		for (i = 0; i < sizeof columnWidths; i++) {
			setColumnWidth(sodius_getRowCell(rowObject, i), columnWidths[i]);
		}
		
		return;
	}
}

void sodius_softDeleteObject(Object o) {
	if (null o) {
		return;
	}
	
	noError;
	softDelete(o);
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot delete object " o."Absolute Number" ": " message);
    }
}

bool sodius_isRTFText(string value) {
    if (null value)
        return false;
        
    if (length(value) > 4) {
        if (value[0] == '{' && value[1] == '\\' && value[2] == 'r' && value[3] == 't' && value[4] == 'f') {
            return true;
        }
    }
        
    return false;
}

void sodius_setAttributeStringValue(Object o, string attributeName, string value) {
	if (null o) {
		return;
	}
	
	noError(); 
    if (sodius_isRTFText(value)) {
        o.attributeName = richText(value);
    }
    else {
        o.attributeName = value;
    }
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set attribute '" attributeName "' of object " o."Absolute Number" ": " message);
    }
}

void sodius_setAttributeIntValue(Object o, string attributeName, int value) {
	if (null o) {
		return;
	}
	
	noError(); 
    o.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set attribute '" attributeName "' of object " o."Absolute Number" ": " message);
    }
}

void sodius_setAttributeRealValue(Object o, string attributeName, real value) {
	if (null o) {
		return;
	}
	
	noError(); 
    o.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set attribute '" attributeName "' of object " o."Absolute Number" ": " message);
    }
}

void sodius_setAttributeBoolValue(Object o, string attributeName, bool value) {
	if (null o) {
		return;
	}
	
	noError(); 
    o.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set attribute '" attributeName "' of object " o."Absolute Number" ": " message);
    }
}

void sodius_setAttributeDateValue(Object o, string attributeName, Date value) {
	if (null o) {
		return;
	}
	
	noError(); 
    o.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set attribute '" attributeName "' of object " o."Absolute Number" ": " message);
    }
}

void sodius_setObjectText(Object o, string value) {
	sodius_setAttributeStringValue(o, "Object Text", value);
}

void sodius_setObjectHeading(Object o, string value) {
	sodius_setAttributeStringValue(o, "Object Heading", value);
}

void sodius_setObjectShortText(Object o, string value) {
	sodius_setAttributeStringValue(o, "Object Short Text", value);
}

void sodius_setModuleAttributeStringValue(string attributeName, string value) {
	noError(); 
    if (sodius_isRTFText(value)) {
        openedModule.attributeName = richText(value);
    }
    else {
        openedModule.attributeName = value;
    }
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set module attribute '" attributeName "': " message);
    }
}

void sodius_setModuleAttributeIntValue(string attributeName, int value) {
	noError(); 
    openedModule.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set module attribute '" attributeName "': " message);
    }
}

void sodius_setModuleAttributeRealValue(string attributeName, real value) {
	noError(); 
    openedModule.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set module attribute '" attributeName "': " message);
    }
}

void sodius_setModuleAttributeBoolValue(string attributeName, bool value) {
	noError(); 
    openedModule.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set module attribute '" attributeName "': " message);
    }
}

void sodius_setModuleAttributeDateValue(string attributeName, Date value) {
	noError(); 
    openedModule.attributeName = value;
    
    string message = lastError();
    if (! null message) {
    	sodius_setError("Cannot set module attribute '" attributeName "': " message);
    }
}

void sodius_setLinkAttributeStringValue(Link l, string attributeName, string value) {
	if (null l) {
		return;
	}
	
	noError(); 
    if (sodius_isRTFText(value)) {
        l.attributeName = richText(value);
    }
    else {
        l.attributeName = value;
    }
    
    string message = lastError();
    if (! null message) {
		ModName_ modname = target(l);
    	sodius_setError("Cannot set attribute '" attributeName "' of link  between Object " sourceAbsNo(l) " and Object " targetAbsNo(l) " (module: " fullName(modname) "): " message);
    }
}

void sodius_setLinkAttributeIntValue(Link l, string attributeName, int value) {
	if (null l) {
		return;
	}
	
	noError(); 
    l.attributeName = value;
    
    string message = lastError();
    if (! null message) {
		ModName_ modname = target(l);
    	sodius_setError("Cannot set attribute '" attributeName "' of link  between Object " sourceAbsNo(l) " and Object " targetAbsNo(l) " (module: " fullName(modname) "): " message);
    }
}

void sodius_setLinkAttributeRealValue(Link l, string attributeName, real value) {
	if (null l) {
		return;
	}
	
	noError(); 
    l.attributeName = value;
    
    string message = lastError();
    if (! null message) {
		ModName_ modname = target(l);
    	sodius_setError("Cannot set attribute '" attributeName "' of link  between Object " sourceAbsNo(l) " and Object " targetAbsNo(l) " (module: " fullName(modname) "): " message);
    }
}

void sodius_setLinkAttributeBoolValue(Link l, string attributeName, bool value) {
	if (null l) {
		return;
	}
	
	noError(); 
    l.attributeName = value;
    
    string message = lastError();
    if (! null message) {
		ModName_ modname = target(l);
    	sodius_setError("Cannot set attribute '" attributeName "' of link  between Object " sourceAbsNo(l) " and Object " targetAbsNo(l) " (module: " fullName(modname) "): " message);
    }
}

void sodius_setLinkAttributeDateValue(Link l, string attributeName, Date value) {
	if (null l) {
		return;
	}
	
	noError(); 
    l.attributeName = value;
    
    string message = lastError();
    if (! null message) {
		ModName_ modname = target(l);
    	sodius_setError("Cannot set attribute '" attributeName "' of link  between Object " sourceAbsNo(l) " and Object " targetAbsNo(l) " (module: " fullName(modname) "): " message);
    }
}

ExternalLink sodius_getExternalLinkByLinkPath(Object o, ExternalLinkDirection d, string linkPath) {
	if (null o) {
		return null;
	}
	
	ExternalLink e;
	if (d == inward) {
		for e in o<-"" do {
			if (linkPath == body(e)) {
				return e;
			}
		}
	}
	else {
		for e in o->"" do {
			if (linkPath == body(e)) {
				return e;
			}
		}
	}
	
	sodius_setError("The path does not match an existing external link of object " o."Absolute Number" ": " linkPath);
	return null;
}

void sodius_createExternalLink(Object o, string n, string desc, ExternalLinkDirection dir, ExternalLinkBehaviour behavior, string linkPath) {
	if (null o) {
		return;
	}
	
	ExternalLink e;
	string message = create(o, desc, n, dir, behavior, linkPath, e); 
	if (! null message) {
    	sodius_setError("Cannot create external link '" n "' of object " o."Absolute Number" ": " message);
		return;
    }
}

void sodius_updateExternalLink(Object o, ExternalLink e, string n, string desc, ExternalLinkDirection dir, ExternalLinkBehaviour behavior, string linkPath) {
	if (null e) {
		return;
	}
	
	string message = update(desc, n, dir, behavior, linkPath, e); 
	if (! null message) {
    	sodius_setError("Cannot update external link '" n "' of object " o."Absolute Number" ": " message);
		return;
    }
}

void sodius_deleteExternalLink(ExternalLink e) {
	if (null e) {
		return;
	}
	
	delete(e);
}

void sodius_readTargetModule(Item i, string variableName) {
	if (null i) {
		return;
	}
	
	Module targetModule = module(i); // already open?
	if (! null targetModule) {
		put(targetModules, variableName, targetModule);
		return;
	}
	
	noError;
	string qualifiedName = fullName(i);
	targetModule = read(qualifiedName, false /* no display */, true /* standard view */);
	lastError; // the module may have DXL errors we are not interested in
	if (null targetModule) {
		sodius_setError("Cannot read module: " qualifiedName);
		return;
	}
	else {
		put(moduleNamesToClose, qualifiedName, qualifiedName);
		put(targetModules, variableName, targetModule);
		return;
	}
}

Module sodius_getTargetModule(string variableName) {
	if (sodius_isTerminated()) {
		return null;
	}
	Module targetModule;
	if (find(targetModules, variableName, targetModule)) {
		return targetModule;
	}
	else {
		sodius_setError("Internal Error, unresolved module variable name: " variableName);
		return null;
	}
}

void sodius_verifyLinkModule(Item i, string variableName) {
	if (null i) {
		return;
	}
	if ("Link" == type(i)) {
		if (! null variableName) {
			sodius_registerLinkModule(variableName, fullName(i));
		}
	}
	else {
		sodius_setError("This DOORS item is not a linked module: " fullName(i));
	}
}

Object sodius_findTargetObject(Module m, int absoluteNumber) {
	if (null m) {
		return null;
	}
	else {
		Object o = object(absoluteNumber, m);
		if (null o) {
			// tables are not acessible with object(absno, module)
			Object tmp;
			for tmp in entire m do {
				int tmpAbsno = tmp."Absolute Number";
				if (tmpAbsno == absoluteNumber) {
					return tmp;
				}
			}
		}
		if (null o) {
			if (type(item(fullName(m))) == "Link") {
				sodius_setError("Invalid command, a link module does contain objects");
				return null;
			}
			else {
				sodius_setError("The absolute number does not match an existing object: " absoluteNumber " in module: " fullName(m)); //DoorsStatusCode
				return null;
			}
		}
		else {
			return o;
		}
	}
}

Object sodius_findTargetObjectByURL(Module m, string url) {
	if (null m) {
		return null;
	}
	
	Item resultItem;
	int absno;
	if (sodius_decodeURL(url, resultItem, absno)) {
		return sodius_findTargetObject(m, absno);
	}
	else if (sodius_decodeResourceURL(url, resultItem, absno)) {
		return sodius_findTargetObject(m, absno);
	}
	else {
		sodius_setError("The URL does not match an existing object: " url " in module: " fullName(m)); //DoorsStatusCode
		return null;
	}
}

void sodius_createLink(Object sourceObject, string linkModuleVariable, Object targetObject) {
	if (null sourceObject || null targetObject) {
		return;
	}
	string linkModuleFullName;
	if (find(linkModulesFullNames, linkModuleVariable, linkModuleFullName)) {
		noError;
		sourceObject -> linkModuleFullName -> targetObject;
		string message = lastError;
		if (! null message) {
			sodius_setError("Cannot create link between Object " sourceObject."Absolute Number" " and Object " targetObject."Absolute Number" " (module: " fullName(module(targetObject)) "): " message);
			return;
		}
	}
	else {
		sodius_setError("Internal Error, unresolved link module variable name: " linkModuleVariable);
	}
}

Link sodius_findLink(Object sourceObject, string linkModuleVariable, Object targetObject) {
	if (null sourceObject || null targetObject) {
		return null;
	}
	string linkModuleFullName;
	if (find(linkModulesFullNames, linkModuleVariable, linkModuleFullName)) {
		Link l = null;
		int absno = targetObject."Absolute Number";
		string fullname = fullName(module(targetObject));
		for l in sourceObject -> linkModuleFullName do {
			if (absno == targetAbsNo(l)) {
				ModName_ targetModName = target(l);
				if (fullname == fullName(targetModName)) {
					return l;
				}
			}
		}
		
		sodius_setError("Cannot find a link between Object " sourceObject."Absolute Number" " and Object " targetObject."Absolute Number" " (module: " fullName(module(targetObject)) ")");
		return null;
	}
	else {
		sodius_setError("Internal Error, unresolved link module variable name: " linkModuleVariable);
		return null;
	}
}

void sodius_deleteLink(Link l) {
	if (null l) {
		return;
	}
	
	noError;
	delete(l);
	string message = lastError;
	if (! null message) {
		ModName_ modname = target(l);
		sodius_setError("Cannot delete link between Object " sourceAbsNo(l) " and Object " targetAbsNo(l) " (module: " fullName(modname) "): " message);
		return;
	}
}

void sodius_createBaseline(Baseline b, string d) {
    if (null openedModule || null b) {
        return;
    }
    noError;
    create(openedModule, b, d);
    string message = lastError();
    if (! null message) {
        sodius_setError("Cannot create baseline " major(b) "." minor(b) " on module '" fullName(openedModule) "': " message);
        return;
    }
}
