
// must be in synch with XMLIdFactory.java
string getIDItem(Item i) {
    return fullName(i);
}

string getIDFolder(Folder f) {
    return fullName(f);
}

string getIDModule(Module m) {
    return fullName(m);
}

string getIDModName(ModName_ m) {
    return fullName(m);
}

string getIDAttrType(Module m, AttrType t) {
    return fullName(m) "#T=" t.name;
}

string getIDAttrDef(Module m, AttrDef a) {
    return fullName(m) "#A=" a.name;
}

string getIDObject(Object o) {
    return fullName(module(o)) "#O=" o."Absolute Number" "";
}

string getIDLinkSet(Object o) {
	string absno = "" o."Absolute Number" "";
	// Linksets do not have an Absolute Number in DOORS 8; need to use the identifier
	if (length(absno) == 0) {
		Module m = module(o);
		string objectId = identifier(o);
		string prefix = m."Prefix";
		if (length(prefix) == 0) {
			absno = objectId;
		}
		else {
			absno = objectId[length(prefix):];
		}
	}
    return fullName(module(o)) "#O=" absno "";
}

string getIDView(Module m, View v) {
    return fullName(m) "#V=" name(v);
}

string getIDBaseline(Module m, Baseline b) {
    string id = fullName(m) "#B=" major(b) "." minor(b) "";
    string s = suffix(b);
    if (! null s)
        return id " (" s ")";
    else
        return id; 
}

string getIDBaseline(ModuleVersion m) {
    return fullName(m) "#B=" versionString(m);
}

// If DOORS client is >= 9.4, there is a DXL function to decode a Resource URL.
string decodeResourceURL_URL = null; 
bool decodeResourceURL_decoded = false; 
Item decodeResourceURL_Item = null; 
int decodeResourceURL_absno = 0; 
string getDecodeResourceURLSourceCode() {
	if (isDoorsClientCompliantWith('9', '4')) {
		return "
		string protocol;
		string dbHost;
		int dbPort;
		string repositoryId;
		string dbName;
		string dbId;
		ModuleVersion modVer;
		string viewName;
		
    	string &decodeResourceURL_URL = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_URL)) "); 
    	bool &decodeResourceURL_decoded = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_decoded)) "); 
    	Item &decodeResourceURL_Item = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_Item)) ");
    	int &decodeResourceURL_absno = addr_(" to32BitLiteral(addr_ (&decodeResourceURL_absno)) ");
	
		string message = null;
		noError;
		message = decodeResourceURL(decodeResourceURL_URL, protocol, dbHost, dbPort, repositoryId, dbName, dbId, 
			decodeResourceURL_Item, modVer, viewName, decodeResourceURL_absno);
		string lastErrorMessage = lastError; 
		decodeResourceURL_decoded = null message && null lastErrorMessage;
		";
	}
	else {
		return "";
	}
}
string decodeResourceURL_sourceCode = getDecodeResourceURLSourceCode();

bool decodeID_ResourceURL(string url, Item &resultItem, int &absno) {
	if (length(decodeResourceURL_sourceCode) == 0) {
		return false;
	}
	else {
		decodeResourceURL_URL = url;
		decodeResourceURL_decoded = false;
		decodeResourceURL_absno = 0;
		
		eval_ decodeResourceURL_sourceCode;
		
		bool result = decodeResourceURL_decoded;
		if (result) {
			resultItem = decodeResourceURL_Item;
			absno = decodeResourceURL_absno;
		}
		
		decodeResourceURL_URL = null;
		decodeResourceURL_decoded = false;
		decodeResourceURL_absno = 0;
		
		return result;
	}
}

bool decodeID_URL(string url, Item &resultItem, int &absno) {
	string dbHost;
	int dbPort;
	string dbName;
	string dbId;
	ModuleVersion modVer;
	
	noError;
	string message = decodeURL(url, dbHost, dbPort, dbName, dbId, resultItem, modVer, absno);
	string lastErrorMessage = lastError;
	bool result = null message && null lastErrorMessage;
	return result;
}

bool decodeID(string id, Item &resultItem, int &absno) {
	
	// might be a DOORS URL
	if (decodeID_URL(id, resultItem, absno)) {
		return true;
	}
	if (decodeID_ResourceURL(id, resultItem, absno)) {
		return true;
	}
	
    string fullName = null;
    int tempAbsno = -1;
    
    int i = 0;
    int l = length(id)
    for (i = 0; i < l; i++) {
        
        // see getID(Object)
        if (id[i] == '#') {
            fullName = id[0 : i - 1];
            if (i + 3 < l) {
                string absnoString = id[i + 3 : l - 1];
                if (isValidInt(absnoString)) {
                    tempAbsno = intOf absnoString;
                }
            } 
        }
    }
    
    if (null fullName) {
        fullName = id;
    }
    
    Item tempItem = item(fullName);
    if (null tempItem) {
    	// The full name might actually be a unique item id
    	tempItem = itemFromID(fullName);
    }
    
    if (! null tempItem) {
        resultItem = tempItem;
        absno = tempAbsno;
        return true;
    }
    else {
        return false;
    }
}

/*
 * Parse baseline in this form: major.minor (suffix)
 */
Baseline parseBaseline(string baselineID) {
    int major = -1;
    int minor = -1;
    string suffix = null;
    
    int i = 0;
    int l = length(baselineID)
    int start = 0
    for (i = 0; i < l; i++) {
        
        // '.' -> previous segment is major
        if (baselineID[i] == '.') {
            major = intOf baselineID[start : i - 1];
            start = i + 1;
        }
        
        // ' ' -> previous segment is minor, next segment is suffix
        else if (baselineID[i] == ' ') {
            if (i + 3 < l && baselineID[i + 1] == '(' && baselineID[l - 1] == ')') {
                minor = intOf baselineID[start : i - 1];
                suffix = baselineID[i + 2 : l - 2];
                return baseline(major, minor, suffix);
            }
            else {
                // weird, unknown pattern
                return null;
            }
        }
    }
    
    // end without suffix?
    if (major != -1 && minor == -1) {
        string minorString = baselineID[start : i - 1];
        if (isValidInt(minorString)) {
            minor = intOf minorString;
            return baseline(major, minor, null);
        }
    }
            
    return null;
}

Item getItemById(string id) {
	
	// might be a full name
    Item i = item(id)
    if (! null i) {
        return i;
	}
	
	// might be an Item ID
	i = itemFromID(id);
	if (! null i) {
		return i;
	}
	
	// might be a DOORS URL
	int absno;
	if (decodeID_URL(id, i, absno)) {
		return i;
	}
	else if (decodeID_ResourceURL(id, i, absno)) {
		return i;
	}
	
	return null;
}

bool findModuleInfoInSkip(Item i, Skip s, string &result) {
    if (find(s, fullName(i), result)) {
    	return true;
    } else if (find(s, uniqueID(i), result)) {
    	return true;
    } else if (find(s, sodius_getURL(i), result)) {
    	return true;
    } else {
    	return false;
    }
}

Module getModule(Item i) {
    
    // baseline?
    string baselineID = null;
    if (findModuleInfoInSkip(i, BASELINE_IDS, baselineID)) {
        Baseline baselineInfo = parseBaseline(baselineID);
        if (! null baselineInfo) {
            
            // already loaded baseline?
            ModName_ modName = module(fullName(i));
            ModuleVersion version = moduleVersion(modName, baselineInfo);
            Module m = data(version);
            if (! null m) {
                return m;
            }
            
            // load it
            else {
                noError(); // bug id=3804: Ignore DXL error that might occur when loading a module 
                Module m = load(version, false);
                lastError();
                return m;
            }
        }
    }
    
    /// read module
    Module m = module(i);
    if (null m) {
    	noError(); // bug id=3804: Ignore DXL error that might occur when loading a module
    	m = read(fullName(i), false);
    	lastError();
    }
    
    // downgrade?
    if (! null m && OPTION_DOWNGRADE_MODULE) {
        if (isEdit(m) || isShare(m)) {
            if (isVisible(m)) {
                if (! unsaved(m)) {
                    downgrade(m);
                }
            }
        }
    }
    
    // use specific view?
    if (! null m) {
	    string expectedViewName = null;
	    if (findModuleInfoInSkip(i, SCOPE_VIEWS, expectedViewName)) {
	    	
	    	// need to switch view?
	    	if (expectedViewName != currentView(m)) {
	    		
	    		// don't close the module if it has unsaved changes
	    		if (! unsaved(m)) {
	    			
	    			// close module
	    			close(m);
	    			
	    			// change default view
	    			string previousDefaultView = getDefaultViewForUser(m);
	    			bool isExpectingStandardView = false;
	    			if (null expectedViewName || expectedViewName == "" || expectedViewName == "Standard view") {
	    				isExpectingStandardView = true;
	    				clearDefaultViewForUser(m);
	    			}
	    			else {
	    				setDefaultViewForUser(m, expectedViewName);
	    			}
	    			
	    			// read again the module
	    			noError(); // bug id=3804: Ignore DXL error that might occur when loading a module
    				m = read(fullName(i), false, isExpectingStandardView);
    				lastError();
    				
    				// reset the old default view
    				if (! null m) {
		    			if (null previousDefaultView || previousDefaultView == "" || previousDefaultView == "Standard view") {
		    				clearDefaultViewForUser(m);
		    			}
		    			else {
		    				setDefaultViewForUser(m, previousDefaultView);
		    			}
    				}
	    		}
	    	}
	    }
    }

    return m; 
}

Module getModule(string id) {
    Item i = getItemById(id)
    if (null i)
        return null;
    else
        return getModule(i);
}

Object internalGetRecursiveObject(int objectAbsno, Object self) {
    int i = self."Absolute Number";
    if (i == objectAbsno) {
        return self;
    }
    
    Object o = null;
    for o in self do {
        Object result = internalGetRecursiveObject(objectAbsno, o);
        if (result != null) {
            return result;
        }
    }
    
    return null;
}

Object internalGetRecursiveObject(int objectAbsno, Module m) {
    Object o = null;
    for o in top(m) do {
        Object result = internalGetRecursiveObject(objectAbsno, o);
        if (! null result) {
            return result;
        }
    }
    
    return null;
}

Object getObject(int objectAbsno, Module m) {
    Object o = object(objectAbsno, m);
                
    // not available -> try to remove the filter applied on the module
    if (null o) {
        if (filtering(m)) {
            current = m;
            filtering off;
            o = object(objectAbsno, m);
            filtering on;
        }
    }
    
    // still not found -> loop on all objects
    // (the object may be a table or a row)
    if (null o) {
        o = internalGetRecursiveObject(objectAbsno, m)
    }
    
    return o;
}

Object getObject(int objectAbsno, Item i) {
    Module m = getModule(i);
    if (! null m) {
        return getObject(objectAbsno, m)
    } 
    
    return null;
}

Object getObject(int objectAbsno, string moduleID) {
    Module m = getModule(moduleID);
    if (! null m) {
        return getObject(objectAbsno, m)
    } 
    
    return null;
}

Object getObject(string id) {
    Item i;
    int objectAbsno = -1;
    if (decodeID(id, i, objectAbsno)) {
        return getObject(objectAbsno, i)
    }
    
    return null;
}

Folder getFolder(string id) {
    Item i = getItemById(id);
    if (! null i) {
        string itemType = type(i);
        if (itemType == "Folder" || itemType == "Project")
            return folder(i);
    }
    
    return null;
}
