bool containsDXLAttributes = false;

void sodius_checkDXLAttribute(AttrDef a) {
	if (!containsDXLAttributes && a.dxl) {
		containsDXLAttributes = true;
	}
}

void sodius_raiseSuccessFlagOnDXLAttributes() {
	if (containsDXLAttributes) {
		error "##SODIUS_SUCCESS_FLAG##";
	}
}

int getScopeMode(Folder folder) {
    if (null folder)
        return OPTION_SCOPE_DEFAULT_MODE;
        
    int mode = 0;
    if (find(SCOPE_MODES, getIDFolder(folder), mode))
        return mode;
        
    mode = getScopeMode(getParentFolder(folder));
    if (mode > OPTION_SCOPE_DEFAULT_MODE)
        return mode;
    else    
        return OPTION_SCOPE_DEFAULT_MODE;
}

int getScopeMode(Item item) {
    if (null item)
        return OPTION_SCOPE_DEFAULT_MODE;
        
    int mode = 0;
    if (find(SCOPE_MODES, getIDItem(item), mode))
        return mode;
        
    mode = getScopeMode(getParentFolder(item));
    if (mode > OPTION_SCOPE_DEFAULT_MODE)
        return mode;
    else    
        return OPTION_SCOPE_DEFAULT_MODE;
}

void cleanup() {
    // delete global Buffers, we do it here unless  
    delete xmlBuffer;
    delete bufferRTF;
    delete bufferXHTML;
    delete bufferOLE;
	//delete bufferValue;  // for some strange reason, this fails in both interactive and batch mode

    delete(SCOPE_MODES);
    delete(SCOPE_VIEWS);
    delete(BASELINE_IDS);
}

int decrementDepth(Skip ids, int depth) {
    if (! null ids) {
        return depth;
    }
    
    if (depth < 0)
        return depth;
    else if (depth == DEPTH_ZERO)
        return DEPTH_ZERO;
    else
        return depth - 1;
}

bool sodius_contains(string v, string fragment, bool matchCase) {
    int offset, l;
    return findPlainText(v, fragment, offset, l, matchCase);
}

bool sodius_includesValue(string multiValue, string v) {
    if (null multiValue) {
        return false;
    }
    if (multiValue == v) {
        return true;
    }
    
    int length = length(multiValue);
    int i = 0, start = 0;
    for (i = 0; i < length; i++) {
        char c = multiValue[i]
        if (multiValue[i] == '\n') {
            if (multiValue[start:i-1] == v) {
                return true;
            }
            start = i + 1;
        }
        else if (i + 1 == length) {
            if (multiValue[start:] == v) {
                return true;
            }
        }
    }
    return false;
}

bool isFeatureRequested(int);
bool isEmptyFeatureRequested(int);

void writeInstantiatedAttribute(string attributeName, Module m) {
    if (null m)
        return;
        
    AttrDef self = find(m, attributeName)
    if (null self) 
        return;
    if (self.system)
        return;
    if (isFilteredAttribute(self))
        return;
	sodius_checkDXLAttribute(self);
        
    string value = m.attributeName
    if (value == null)
        return;
        
    string typeName = stringOf(self.type.type);
    
    // start
    writeXMLStartTag(START_TAG_INSTANTIATEDATTRIBUTE);
    writeXMLAttributeString(ATTR_INSTANTIATEDATTRIBUTE_INSTANTIATES, getIDAttrDef(m, self))
    
    // date?
    if (typeName == "Date") {
        Date d = m.attributeName
        writeXMLAttributeDateAsString(ATTR_INSTANTIATEDATTRIBUTE_VALUE, d)
    }
    
    // enumeration?
    else if (typeName == "Enumeration") {
        writeXMLAttributeEnumeration(ATTR_INSTANTIATEDATTRIBUTE_VALUE, value)
    }
    
    // may be rich text
    else {
        writeXMLAttributeRichText(ATTR_INSTANTIATEDATTRIBUTE_VALUE, m, attributeName)
    }
    
    // end
    writeXMLEndTag();
}

bool isFilteredInstantiatedAttribute(string attributeName, Object o, Module m) {
    if (null o || null m)
        return true;
    if (attributeName == "RTF Pict")
        return true;
    if (attributeName == "RTF NonShpPict")
        return true;
    if (attributeName == "RTF ShpPict")
        return true;
        
    AttrDef self = find(m, attributeName)
    if (null self) 
        return true;
    if (self.system)
        return true;
    if (isFilteredAttribute(self))
        return true;
        
    return false;
}    

void writeInstantiatedAttribute(string attributeName, Object o, Module m) {
    if (isFilteredInstantiatedAttribute(attributeName, o, m))
        return;
        
    string value = o.attributeName
    if (value == null)
        return;
        
    AttrDef self = find(m, attributeName);
    string typeName = stringOf(self.type.type);
	sodius_checkDXLAttribute(self);
    
    // start
    writeXMLStartTag(START_TAG_INSTANTIATEDATTRIBUTE);
    writeXMLAttributeString(ATTR_INSTANTIATEDATTRIBUTE_INSTANTIATES, getIDAttrDef(m, self))
    
    // date?
    if (typeName == "Date") {
        Date d = o.attributeName
        writeXMLAttributeDateAsString(ATTR_INSTANTIATEDATTRIBUTE_VALUE, d)
    }
    
    // enumeration?
    else if (typeName == "Enumeration") {
        writeXMLAttributeEnumeration(ATTR_INSTANTIATEDATTRIBUTE_VALUE, value)
    }
    
    // may be rich text
    else {
        writeXMLAttributeRichText(ATTR_INSTANTIATEDATTRIBUTE_VALUE, o, attributeName)
    }
    
    // end
    writeXMLEndTag();
}

void writeObjectInstantiatedAttributesEmpty(Object self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES))
        return;
    
    string attributeName
    Module m = module(self);
    for attributeName in m do {
        if (! isFilteredInstantiatedAttribute(attributeName, self, m))
            return;
    }
    
    writeXMLStartTag(START_TAG_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES);
    writeXMLEndTag();
}

void writeObjectInstantiatedAttributes(Object self) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES)) {
        writeObjectInstantiatedAttributesEmpty(self);
        return;
    }
    
    writeXMLStartTag(START_TAG_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES);
    string attributeName
    Module m = module(self);
    for attributeName in m do {
        writeInstantiatedAttribute(attributeName, self, m);
    }
    writeXMLEndTag();
}

void writeObjectImageEmpty(Object self) {
    if (null self) 
        return;
    if (OPTION_IGNORE_IMAGE)
        return;
    if (! isEmptyFeatureRequested(FEATURE_OBJECT_IMAGE))
        return;
        
    if (! oleIsObject(self) && getPictFormat(self) == 0)
        return;
    
    writeXMLAttributeString(ATTR_OBJECT_IMAGE, "true");
}

void writeObjectImage(Object self) {
    if (null self) 
        return;
    if (OPTION_IGNORE_IMAGE)
        return;
    if (! isFeatureRequested(FEATURE_OBJECT_IMAGE)) {
        writeObjectImageEmpty(self);
        return;
    }
    if (! OPTION_IGNORE_OLE_PREVIEW_IMAGES && getPictFormat(self) == 0) {
    	return;
    }

    string imageFilename = createImageFilename(formatPNG);
    if (null exportPicture(self, imageFilename, formatPNG)) {
        writeXMLAttributeString(ATTR_OBJECT_IMAGE, imageFilename);
    }
}

void writeExternalLink(ExternalLink self, string direction) {
    if (null self)
        return;
        
    Date d;
    writeXMLStartTag(START_TAG_EXTERNALLINK);
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, name(self));
    writeXMLAttributeString(ATTR_EXTERNALLINK_LINKPATH, body(self));
    writeXMLAttributeString(ATTR_EXTERNALLINK_DIRECTION, direction);
    writeXMLAttributeString(ATTR_EXTERNALLINK_OPERATION, behaviour(self) == openAsURL ? "openAsURL" : "none");
    writeXMLAttributeString(ATTR_EXTERNALLINK_DESCRIPTION, description(self));
    writeXMLAttributeString(ATTR_EXTERNALLINK_CREATEDBY, self."Created By" "");
    d = self."Created On";
    writeXMLAttributeDate(ATTR_EXTERNALLINK_CREATEDON, d);
    writeXMLAttributeString(ATTR_EXTERNALLINK_LASTMODIFIEDBY, self."Last Modified By" "");
    d = self."Last Modified On";
    writeXMLAttributeDate(ATTR_EXTERNALLINK_LASTMODIFIEDON, d);
    writeXMLEndTag();
}

void writeObjectExternalLinksEmpty(Object self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_OBJECT_EXTERNALLINKS))
        return;
    
    ExternalLink e;
    for e in self -> "" do {
        return;
    }
    for e in self <- "" do {
        return;
    }
    
    writeXMLStartTag(START_TAG_OBJECT_EXTERNALLINKS);
    writeXMLEndTag();
}

void writeObjectExternalLinks(Object self) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_OBJECT_EXTERNALLINKS)) {
        writeObjectExternalLinksEmpty(self);
        return;
    }
    
    writeXMLStartTag(START_TAG_OBJECT_EXTERNALLINKS);
    ExternalLink e;
    for e in self -> "" do {
        writeExternalLink(e, "outward");
    }
    for e in self <- "" do {
        writeExternalLink(e, "inward");
    }
    writeXMLEndTag();
}

void writeObjectDiscussionsEmpty(Object);
void writeObjectDiscussions(Object);

void writeObject(Object, Skip, int);

Date getLastModifiedOn(Object self) {
    Date d = null, latest = null;
    if (OPTION_PRECISE_LAST_MODIFIED_ON) {
        History h;
        for h in self do {
            d = h.date; if (latest == null || latest < d) { latest = d };
        } 
    }
    d = lastModifiedTime(self); if (latest == null || latest < d) { latest = d };
    if (null d) {
        d = self."Last Modified On"; if (latest == null || latest < d) { latest = d };
    }
    return latest;
}

void writeObjectSystemAttributes(Object self) {
    if (null self) 
        return;
        
    writeXMLAttributeString(ATTR_ATTRIBUTEDELEMENT_CREATEDBY, self."Created By" "");
    Date d = self."Created On";
    writeXMLAttributeDate(ATTR_ATTRIBUTEDELEMENT_CREATEDON, d);
    writeXMLAttributeString(ATTR_ATTRIBUTEDELEMENT_LASTMODIFIEDBY, self."Last Modified By" "");
    d = getLastModifiedOn(self);
    writeXMLAttributeDate(ATTR_ATTRIBUTEDELEMENT_LASTMODIFIEDON, d);
    writeXMLAttributeInt(ATTR_OBJECT_OBJECTLEVEL, level(self));
    writeXMLAttributeRichText(ATTR_OBJECT_OBJECTHEADING, self, "Object Heading");
    writeXMLAttributeRichText(ATTR_OBJECT_OBJECTTEXT, self, "Object Text");
    writeXMLAttributeRichText(ATTR_OBJECT_OBJECTSHORTTEXT, self, "Object Short Text");
    writeXMLAttributeString(ATTR_OBJECT_OBJECTNUMBER, number(self));
    writeXMLAttributeString(ATTR_OBJECT_OBJECTIDENTIFIER, identifier(self));
}

void writeObjectObjectsEmpty(Object self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_OBJECT_NESTEDOBJECTS))
        return;
        
    Object o
    for o in self do {
        return;
    }
    
    writeXMLStartTag(START_TAG_OBJECT_NESTEDOBJECTS);
    writeXMLEndTag();
}

void writeObjectObjects(Object self, Skip ids, int depth) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_OBJECT_NESTEDOBJECTS)) {
        writeObjectObjectsEmpty(self);
        return;
    }
    
    current = module(self);
    bool showDeleted = showDeletedObjects();
    if (showDeleted != OPTION_SHOW_DELETED_OBJECTS)
        showDeletedObjects(OPTION_SHOW_DELETED_OBJECTS);
        
    writeXMLStartTag(START_TAG_OBJECT_NESTEDOBJECTS);
    Object o
    for o in self do {
        checkCanceled();
        writeObject(o, ids, depth);
    }
    writeXMLEndTag();
    
    if (showDeleted != OPTION_SHOW_DELETED_OBJECTS)
        showDeletedObjects(showDeleted);
}

void writeObjectChildren(Object self, Skip ids, int depth) {
    writeObjectSystemAttributes(self);
    writeObjectImage(self);
    writeObjectInstantiatedAttributes(self);
    writeObjectExternalLinks(self);
    writeObjectDiscussions(self);
    writeObjectObjects(self, ids, depth);
}

void writeObjectChildrenEmpty(Object self) {
    writeObjectImageEmpty(self);
    writeObjectInstantiatedAttributesEmpty(self);
    writeObjectExternalLinksEmpty(self);
    writeObjectDiscussionsEmpty(self);
    writeObjectObjectsEmpty(self);
}

bool isObjectChildrenRequested(Object self, Skip ids, int depth) {
    checkCanceled();
    
    if (! null ids) {
        int mode = 0;
        return find(ids, getIDObject(self), mode);
    }
    else if (depth != DEPTH_ZERO) {
        return isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_CREATEDBY) || 
            isFeatureRequested(FEATURE_OBJECT_IMAGE) || 
            isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES) || 
            isFeatureRequested(FEATURE_OBJECT_EXTERNALLINKS) || 
            isFeatureRequested(FEATURE_OBJECT_NESTEDOBJECTS) ||
            isFeatureRequested(FEATURE_DISCUSSIONCONTAINER_DISCUSSIONS)
    }
    else {
        return false;
    }
}

void writeObjectTableOrRow(Object self, Skip ids, int depth, string objectKind) {
    
    // interested in Table or Row object?
    if (! OPTION_IGNORE_TABLE) {
        writeXMLStartTag(START_TAG_OBJECT);
        writeXMLAttributeString(ATTR_SITUATEDELEMENT_URL, sodius_getURL(self));
        writeXMLAttributeRichText(ATTR_OBJECT_OBJECTHEADING, self, "Object Heading");
        writeXMLAttributeString(ATTR_OBJECT_IMAGE, "false");
        writeXMLAttributeString(ATTR_OBJECT_OBJECTKIND, objectKind);
        if (isDeleted(self)) {
            writeXMLAttributeBool(ATTR_OBJECT_ISDELETED, true);
        }
        int i = self."Absolute Number"
    	writeXMLAttributeInt(ATTR_OBJECT_ABSOLUTENUMBER, i);
    	writeXMLAttributeString(ATTR_OBJECT_OBJECTNUMBER, number(self));
        writeXMLAttributeString(ATTR_OBJECT_OBJECTIDENTIFIER, identifier(self));
        
        writeXMLStartTag(START_TAG_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES);
        writeXMLEndTag();
        writeXMLStartTag(START_TAG_OBJECT_EXTERNALLINKS);
        writeXMLEndTag();

        writeObjectObjects(self, ids, depth);

        writeXMLEndTag();
    }
    
    else {
        // just write children
        Object o
        for o in self do {
            writeObject(o, ids, depth);
        }
    }
}

void writeObject(Object self, Skip ids, int depth) {
    if (null self) 
        return;
     
    // table?   
    if (table(self)) {
        writeObjectTableOrRow(self, ids, depth, "table");
        return;
    }
    if (row(self)) {
        writeObjectTableOrRow(self, ids, depth, "row");
        return;
    }
    
    // start
    writeXMLStartTag(START_TAG_OBJECT);
    writeXMLAttributeString(ATTR_SITUATEDELEMENT_URL, sodius_getURL(self));
    int i = self."Absolute Number"
    writeXMLAttributeInt(ATTR_OBJECT_ABSOLUTENUMBER, i);

    string createdThru = self."Created Thru" "";
    if (createdThru != "Manual Input") { // default value
        writeXMLAttributeString(ATTR_OBJECT_CREATEDTHRU, createdThru);
    }

    if (cell(self)) {
        writeXMLAttributeString(ATTR_OBJECT_OBJECTKIND, "cell");
    }
    if (isDeleted(self)) {
        writeXMLAttributeBool(ATTR_OBJECT_ISDELETED, true);
    }
    writeXMLStartTagEnd();
    
    // children
    if (isObjectChildrenRequested(self, ids, depth)) {
        writeObjectChildren(self, ids, decrementDepth(ids, depth));
    }
    else {
        // not resolving and not interested in children? We might still want to get the system attributes
        // which were not considered as children in the first place (MMDOORS-96)
        if (null ids && depth == DEPTH_ZERO && isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_CREATEDBY)) {
            writeObjectSystemAttributes(self);
        }
        
        writeObjectChildrenEmpty(self);
    }
    
    // end
    writeXMLEndTag();
}

void writeColumn(Column self, Module m) {
    if (null self)
        return;
        
    writeXMLStartTag(START_TAG_COLUMN);
    writeXMLAttributeString(ATTR_COLUMN_TITLE, title(self))
    writeXMLAttributeString(ATTR_COLUMN_JUSTIFICATION, lower(justify(self)) == "left and right" ? "full" : lower(justify(self)));
    writeXMLAttributeInt(ATTR_COLUMN_WIDTH, width(self))
    
    string attributeName = attrName(self);
    if (attributeName != null) {
        if (attributeName == "Object Level")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "ObjectLevel");
        else if (attributeName == "Object Identifier")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "ObjectIdentifier");
        else if (attributeName == "Object Number")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "ObjectNumber");
        else if (attributeName == "Absolute Number")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "AbsoluteNumber");
        else if (attributeName == "Created By")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "CreatedBy");
        else if (attributeName == "Created On")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "CreatedOn");
        else if (attributeName == "Created Thru")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "CreatedThru");
        else if (attributeName == "Last Modified By")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "LastModifiedBy");
        else if (attributeName == "Last Modified On")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "LastModifiedOn");
        else if (attributeName == "Object Heading")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "ObjectHeading");
        else if (attributeName == "Object Short Text")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "ObjectShortText");
        else if (attributeName == "Object Text")
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "ObjectText");
        else {
            AttrDef attr = find(m, attributeName)
            if (attr != null && ! isFilteredAttribute(attr)) {
                writeXMLAttributeString(ATTR_COLUMN_USERDEFINEDATTRIBUTE, getIDAttrDef(m, attr));
            }
        }
    }
    else {
        // dxl?
        string dxlCode = dxl(self);
        if (dxlCode != null) {
            writeXMLAttributeString(ATTR_COLUMN_DXL, dxlCode);
        }
        else {
            // attribute is objectHeading and objecttext
            writeXMLAttributeString(ATTR_COLUMN_BASEATTRIBUTE, "ObjectHeadingAndObjectText");
        }
    }
    
    writeXMLEndTag();
}

void writeView(View self, Module m) {
    if (null self)
        return;
        
    ViewDef vd = get(m, self)
    string name = name(self);
    
    writeXMLStartTag(START_TAG_VIEW);
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, name);
    writeXMLAttributeBool(ATTR_VIEW_USECURRENT, useCurrent(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEWINDOWS, useWindows(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESORTING, useSorting(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEFILTERING, useFiltering(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEANCESTORS, useAncestors(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEDESCENDANTS, useDescendants(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESELECTION, useSelection(vd));
    writeXMLAttributeBool(ATTR_VIEW_USECOLUMNS, useColumns(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEFILTERTABLES, useFilterTables(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEGRAPHICSCOLUMN, useGraphicsColumn(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESHOWEXPLORER, useShowExplorer(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEGRAPHICS, useGraphics(vd));
    writeXMLAttributeBool(ATTR_VIEW_USEOUTLINING, useOutlining(vd));
    writeXMLAttributeBool(ATTR_VIEW_USECOMPRESSION, useCompression(vd));
    writeXMLAttributeBool(ATTR_VIEW_USELEVEL, useLevel(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESHOWDELETED, useShowDeleted(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESHOWPICTURES, useShowPictures(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESHOWTABLES, useShowTables(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESHOWLINKINDICATORS, useShowLinkIndicators(vd));
    writeXMLAttributeBool(ATTR_VIEW_USESHOWLINKS, useShowLinks(vd));
    writeXMLAttributeBool(ATTR_VIEW_USETOOLTIPCOLUMN, useTooltipColumn(vd));
    
    // columns
    if (useColumns(vd)) {
        
        // Need to ensure this view is the default one for the module, otherwise the columns would be inaccurate
        if (name == currentView(m)) {
            writeXMLStartTag(START_TAG_VIEW_COLUMNS);
            Column c;
            for c in m do {
                writeColumn(c, m);
            }
            writeXMLEndTag();
        }
    }
    
    writeXMLEndTag();
}

void writeFormalModuleViewsEmpty(Module self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_FORMALMODULE_VIEWS))
        return;
    
    current = self; 
    string name;
    for name in views(self) do {
        return;
    }
    
    writeXMLStartTag(START_TAG_FORMALMODULE_VIEWS);
    writeXMLEndTag();
}

bool canChangeDefaultView(Module self, string viewName) {
    if (OPTION_IGNORE_NON_CURRENT_VIEW_COLUMNS)
        return false;

    if (isVisible(self))
        return false;
        
    // [MMDOORS-33] Empty view columns for baselined module
    if (isBaseline(self)) {
    	return true;
    }
        
    if (viewName == currentView(self))
        return false;
        
    return true;        
}

void writeFormalModuleViews(Module &self) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_FORMALMODULE_VIEWS)) {
        writeFormalModuleViewsEmpty(self);
        return;
    }
    
    // views
    current = self; 
    string originalCurrentView = currentView(self);
    string originalDefaultViewName = getDefaultViewForUser(self);
    string moduleFullName = fullName(self);

    writeXMLStartTag(START_TAG_FORMALMODULE_VIEWS);
    string name;
    View v;
    bool wasBaseline = isBaseline(self);
    for name in views(self) do {
        
        // need to change the default view?
        if (canChangeDefaultView(self, name)) {
            setDefaultViewForUser(self, name);
            close(self);
            
            // Note: do not use getModule(moduleFullName), we want the current state, not a potential baseline
            // bug id=3807: Can't get columns of views when reading a baseline 
            noError(); // bug id=3804: Ignore DXL error that might occur when loading a module 
            self = read(moduleFullName, false);
            lastError();
        }
        
        // write view
        if (! null self) {
            writeView(view(name), self);
        }
    }
    writeXMLEndTag();
    
    // need to restore the view?
    if (! OPTION_IGNORE_NON_CURRENT_VIEW_COLUMNS && (wasBaseline || originalCurrentView != currentView(self))) {
        if (originalDefaultViewName != null && originalDefaultViewName != "")
            setDefaultViewForUser(self, originalDefaultViewName);
        else
            clearDefaultViewForUser(self);
            
        close(self);
        self = getModule(moduleFullName);
    }
    
    // current view
    name = currentView(self);
    if (! null name && length(name) != 0 && name != "Standard view") {
        View view = view(name);
        if (! null view) {
            writeXMLAttributeString(ATTR_FORMALMODULE_CURRENTVIEW, getIDView(self, view));
        }
        else {
            writeXMLAttributeStringEmpty(ATTR_FORMALMODULE_CURRENTVIEW);
        }
    }
    else {
        writeXMLAttributeStringEmpty(ATTR_FORMALMODULE_CURRENTVIEW);
    }
}

void writeBaseline(Baseline self) {
    if (null self)
        return;
        
    writeXMLStartTag(START_TAG_BASELINE);
    writeXMLAttributeInt(ATTR_BASELINEINFO_MAJOR, major(self));
    writeXMLAttributeInt(ATTR_BASELINEINFO_MINOR, minor(self));
    writeXMLAttributeString(ATTR_BASELINEINFO_SUFFIX, suffix(self));
    writeXMLAttributeString(ATTR_BASELINEINFO_DESCRIPTION, annotation(self));
    writeXMLAttributeString(ATTR_BASELINEINFO_CREATEDBY, user(self));
    writeXMLAttributeDate(ATTR_BASELINEINFO_CREATEDON, dateOf(self));
    writeXMLEndTag();
}

void writeFormalModuleBaselinesEmpty(Module self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_FORMALMODULE_BASELINES))
        return;
    
    current = self; 
    Baseline b;
    for b in self do {
        return;
    }
    
    writeXMLStartTag(START_TAG_FORMALMODULE_BASELINES);
    writeXMLEndTag();
}

void writeFormalModuleBaselines(Module self) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_FORMALMODULE_BASELINES)) {
        writeFormalModuleBaselinesEmpty(self);
        return;
    }
    
    // baselines
    current = self; 
    writeXMLStartTag(START_TAG_FORMALMODULE_BASELINES);
    Baseline b;
    for b in self do {
        writeBaseline(b);
    }
    writeXMLEndTag();
    
    // current baseline
    if (isBaseline(self)) {
        writeXMLAttributeString(ATTR_FORMALMODULE_BASELINE, getIDBaseline(self, baselineInfo(self)));
    }
}

void writeFormalModuleDiscussions(Module);

void writeType(AttrType self) {
    if (null self)
        return;
       
    string typeName = self.type "";

    // start
    writeXMLStartTag(START_TAG_TYPE);
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, self.name)
    writeXMLAttributeBool(ATTR_TYPE_SYSTEM, self.system)
    writeXMLAttributeString(ATTR_TYPE_BASETYPE, typeName);
    
    // int or real?
    if (typeName == "Integer") {
        noError
        if (self.minValue) {
        	int i = self.minValue;
            writeXMLAttributeString(ATTR_TYPE_MINVALUE, i "");
        }
        if (self.maxValue) { 
        	int i = self.maxValue;
            writeXMLAttributeString(ATTR_TYPE_MAXVALUE, i "");
        }
        lastError
    }
    
    // real?
    else if (typeName == "Real") {
        noError
        if (self.minValue) {
        	real i = self.minValue;
            writeXMLAttributeString(ATTR_TYPE_MINVALUE, i "");
        }
        if (self.maxValue) { 
        	real i = self.maxValue;
            writeXMLAttributeString(ATTR_TYPE_MAXVALUE, i "");
        }
        lastError
    }
    
    // date?
    else if (typeName == "Date") {
        noError
        if (self.minValue) 
            writeXMLAttributeDateAsString(ATTR_TYPE_MINVALUE, self.minValue);
        if (self.maxValue) 
            writeXMLAttributeDateAsString(ATTR_TYPE_MAXVALUE, self.maxValue);
        lastError
    }

    // enumeration?
    else if (typeName == "Enumeration") {
        writeXMLStartTag(START_TAG_TYPE_ENUMERATIONLITERALS);
        int i;
        for (i = 0; i < self.size; i++) {
            writeXMLStartTag(START_TAG_ENUMERATIONLITERAL);
            writeXMLAttributeString(ATTR_ENUMERATIONLITERAL_VALUE, self.strings[i]);
            writeXMLAttributeInt(ATTR_ENUMERATIONLITERAL_RELATEDNUMBER, self.values[i]);
            writeXMLEndTag();
        }
        writeXMLEndTag();
    }

    writeXMLEndTag();
}

void writeAttribute(AttrDef self, Module m) {
    if (null self)
        return;
    if (self.system && OPTION_IGNORE_SYSTEM_ATTRIBUTES)
        return;
    if (isFilteredAttribute(self))
        return;
            
    writeXMLStartTag(START_TAG_ATTRIBUTE);
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, self.name);
    writeXMLAttributeString(ATTR_ATTRIBUTE_TYPE, getIDAttrType(m, self.type));
    writeXMLAttributeBool(ATTR_ATTRIBUTE_SYSTEM, self.system);
    writeXMLAttributeBool(ATTR_ATTRIBUTE_EXISTSFOROBJECTS, self.object);
    writeXMLAttributeBool(ATTR_ATTRIBUTE_EXISTSFORMODULE, self.module);
    writeXMLAttributeBool(ATTR_ATTRIBUTE_INHERITVALUE, self.inherit);
    writeXMLAttributeBool(ATTR_ATTRIBUTE_AFFECTCHANGEBARS, ! self.nobars);
    writeXMLAttributeBool(ATTR_ATTRIBUTE_AFFECTCHANGEDATES, ! self.nochanges);
    writeXMLAttributeBool(ATTR_ATTRIBUTE_GENERATEHISTORY, ! self.nohistory);
    writeXMLAttributeBool(ATTR_ATTRIBUTE_MULTIVALUED, self.multi);
    string s = self.dxl;
    writeXMLAttributeString(ATTR_ATTRIBUTE_DXL, s);
    
    // default value
    if (self.defval) {
        string typeName = stringOf(self.type.type)
        if (typeName == "Date") {
            Date d = self.defval;
            writeXMLAttributeDateAsString(ATTR_ATTRIBUTE_DEFAULTVALUE, d);
        }
        else if (typeName == "Enumeration") {
            writeXMLAttributeEnumeration(ATTR_ATTRIBUTE_DEFAULTVALUE, self.defval);
        }
        else {
            writeXMLAttributeString(ATTR_ATTRIBUTE_DEFAULTVALUE, self.defval);
        }
    }

    writeXMLEndTag();
}
       
void writeModuleSystemAttributes(Module self) {
    if (null self)
        return;
    if (! isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_CREATEDBY))
        return;
       
    // system attributes 
    Date d;
    writeXMLAttributeString(ATTR_ATTRIBUTEDELEMENT_CREATEDBY, self."Created By" "")
    d = self."Created On"
    writeXMLAttributeDate(ATTR_ATTRIBUTEDELEMENT_CREATEDON, d)
    writeXMLAttributeString(ATTR_ATTRIBUTEDELEMENT_LASTMODIFIEDBY, self."Last Modified By" "")
    d = lastModifiedTime(self); if (null d) { d = self."Last Modified On" }
    writeXMLAttributeDate(ATTR_ATTRIBUTEDELEMENT_LASTMODIFIEDON, d)
    writeXMLAttributeString(ATTR_MODULE_PREFIX, self."Prefix" "")
    
    if (! isVisible(self))
        writeXMLAttributeString(ATTR_MODULE_MODE, "close");
    else if (isShare(self))
        writeXMLAttributeString(ATTR_MODULE_MODE, "share");
    else if (isEdit(self))
        writeXMLAttributeString(ATTR_MODULE_MODE, "edit");
    else
        writeXMLAttributeString(ATTR_MODULE_MODE, "read");
}
 
void writeModuleTypesEmpty(Module self) {
    if (null self)
        return;
    if (! isEmptyFeatureRequested(FEATURE_MODULE_TYPES))
        return;

    AttrType attrType;
    for attrType in self do {
        return;
    }

    writeXMLStartTag(START_TAG_MODULE_TYPES);
    writeXMLEndTag();
}
    
void writeModuleTypes(Module self) {
    if (null self)
        return;
    if (! isFeatureRequested(FEATURE_MODULE_TYPES)) {
        writeModuleTypesEmpty(self);
        return;
    }

    writeXMLStartTag(START_TAG_MODULE_TYPES);
    AttrType attrType;
    for attrType in self do {
        writeType(attrType);
    }
    writeXMLEndTag();
}
    
void writeModuleAttributesEmpty(Module self) {
    if (null self)
        return;
    if (! isEmptyFeatureRequested(FEATURE_MODULE_ATTRIBUTES))
        return;

    AttrDef attr;
    for attr in self do {
        if (! isFilteredAttribute(attr))
            return;
    }
    
    writeXMLStartTag(START_TAG_MODULE_ATTRIBUTES);
    writeXMLEndTag();
}
    
void writeModuleAttributes(Module self) {
    if (null self)
        return;
    if (! isFeatureRequested(FEATURE_MODULE_ATTRIBUTES)) {
        writeModuleAttributesEmpty(self);
        return;
    }

    writeXMLStartTag(START_TAG_MODULE_ATTRIBUTES);
    AttrDef attr;
    for attr in self do {
        writeAttribute(attr, self);
    }
    writeXMLEndTag();
}
    
void writeModuleInstantiatedAttributesEmpty(Module self) {
    if (null self)
        return;
    if (! isEmptyFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES))
        return;
    
    string attributeName
    for attributeName in attributes(self) do {
        return;
    }
    
    writeXMLStartTag(START_TAG_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES);
    writeXMLEndTag();
}

void writeModuleInstantiatedAttributes(Module self) {
    if (null self)
        return;
    if (! isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES)) {
        writeModuleInstantiatedAttributesEmpty(self);
        return;
    }
    
    writeXMLStartTag(START_TAG_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES);
    string attributeName
    for attributeName in attributes(self) do {
        writeInstantiatedAttribute(attributeName, self);
    }
    writeXMLEndTag();
}

void writeModuleItemSystemAttributes(Item self) {
    if (null self)
        return;
        
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, name(self))
    writeXMLAttributeString(ATTR_SITUATEDELEMENT_URL, sodius_getURL(self))
    writeXMLAttributeString(ATTR_ITEM_DESCRIPTION, description(self))
}

void writeFormalModuleSystemAttributes(Module self) {
    if (null self) 
        return;
    
    writeXMLAttributeString(ATTR_SITUATEDELEMENT_URL, sodius_getURL(self))
    writeModuleSystemAttributes(self);
}
        
void writeFormalModuleObjectsEmpty(Module self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_FORMALMODULE_OBJECTS))
        return;
    
    Object o
    for o in top(self) do {
        return;
    }

    writeXMLStartTag(START_TAG_FORMALMODULE_OBJECTS);
    writeXMLEndTag();
}

void writeFormalModuleObjects(Module self, Skip ids, int depth) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_FORMALMODULE_OBJECTS)) {
        writeFormalModuleObjectsEmpty(self);
        return;
    }
    
    current = self;
    bool showDeleted = showDeletedObjects();
    if (showDeleted != OPTION_SHOW_DELETED_OBJECTS)
        showDeletedObjects(OPTION_SHOW_DELETED_OBJECTS);
        
    writeXMLStartTag(START_TAG_FORMALMODULE_OBJECTS);
    Object o
    for o in top(self) do {
        checkCanceled();
        writeObject(o, ids, depth);
    }
    writeXMLEndTag();
    
    if (showDeleted != OPTION_SHOW_DELETED_OBJECTS)
        showDeletedObjects(showDeleted);
}
        
void writeFormalModuleViewObjectsEmpty(Module self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_FORMALMODULE_VIEWOBJECTS))
        return;
    
    Object o
    for o in self do {
        return;
    }

    writeXMLAttributeStringEmpty(ATTR_FORMALMODULE_VIEWOBJECTS);
}

void writeFormalModuleViewObjectsImpl(Module);

void writeFormalModuleViewObjectsDefaultImpl(Module self) {
    Object o
    for o in self do {
        writeXMLAttributeString(ATTR_FORMALMODULE_VIEWOBJECTS, getIDObject(o))
    }
}

void writeFormalModuleViewObjects(Module self) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_FORMALMODULE_VIEWOBJECTS)) {
        writeFormalModuleViewObjectsEmpty(self);
        return;
    }
    
    current = self;
    bool showDeleted = showDeletedObjects();
    if (showDeleted != OPTION_SHOW_DELETED_OBJECTS)
        showDeletedObjects(OPTION_SHOW_DELETED_OBJECTS);
        
    writeFormalModuleViewObjectsImpl(self);
    
    if (showDeleted != OPTION_SHOW_DELETED_OBJECTS)
        showDeletedObjects(showDeleted);
}

void writeFormalModuleChildren(Module self, Skip ids, int depth) {
    writeFormalModuleSystemAttributes(self);
    writeModuleTypes(self);
    writeModuleAttributes(self);
    writeModuleInstantiatedAttributes(self);
    writeFormalModuleObjects(self, ids, depth);
    writeFormalModuleViewObjects(self);
    writeFormalModuleViews(self);
    writeFormalModuleBaselines(self);
    writeFormalModuleDiscussions(self);
}

bool isFormalModuleChildrenRequested(Item self, Skip ids, int depth) {
    checkCanceled();
    
    if (! null ids) {
        int mode = 0;
        return find(ids, getIDItem(self), mode);
    }
    else if (depth != DEPTH_ZERO) {
        return isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_CREATEDBY) || 
            isFeatureRequested(FEATURE_MODULE_TYPES) || 
            isFeatureRequested(FEATURE_MODULE_ATTRIBUTES) || 
            isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES) || 
            isFeatureRequested(FEATURE_FORMALMODULE_OBJECTS) ||
            isFeatureRequested(FEATURE_FORMALMODULE_VIEWOBJECTS) ||
            isFeatureRequested(FEATURE_FORMALMODULE_VIEWS) ||
            isFeatureRequested(FEATURE_FORMALMODULE_BASELINES) ||
            isFeatureRequested(FEATURE_DISCUSSIONCONTAINER_DISCUSSIONS);
    }
    else {
        return false;
    }
}

bool isExcludedFormalModule(Item self, Skip ids, int depth) {
    if (isFilteredFormalModule(self))
        return true;
    else 
        return getScopeMode(self) == SCOPE_MODE_EXCLUDE;
}

void writeFormalModule(Item self, Skip ids, int depth) {
    if (null self) 
        return;
    if (isExcludedFormalModule(self, ids, depth))
        return;
    
    // start
    writeXMLStartTag(START_TAG_FORMALMODULE);
    writeModuleItemSystemAttributes(self);
    
    // children
    if (isFormalModuleChildrenRequested(self, ids, depth)) {
        if (getScopeMode(self) == SCOPE_MODE_COMPLETE)
            writeFormalModuleChildren(getModule(self), null, DEPTH_LOCAL_INFINITE);
        else
            writeFormalModuleChildren(getModule(self), ids, decrementDepth(ids, depth));
    }
    else {
        // nothing, we don't want to load the module
    }
    
    // end
    writeXMLEndTag();
}

void writeInstantiatedAttribute(string attributeName, Link l, Module m) {
    if (null l || null m)
        return;
    if (attributeName == "RTF Pict")
        return;
    if (attributeName == "RTF NonShpPict")
        return true;
    if (attributeName == "RTF ShpPict")
        return true;
        
    AttrDef self = find(m, attributeName)
    if (null self) 
        return;
    if (self.system)
        return;
    if (isFilteredAttribute(self))
        return;
	sodius_checkDXLAttribute(self);
        
    string value = l.attributeName
    if (value == null)
        return;
        
    string typeName = stringOf(self.type.type);
    
    // start
    writeXMLStartTag(START_TAG_INSTANTIATEDATTRIBUTE);
    writeXMLAttributeString(ATTR_INSTANTIATEDATTRIBUTE_INSTANTIATES, getIDAttrDef(m, self))
    
    // date?
    if (typeName == "Date") {
        Date d = l.attributeName
        writeXMLAttributeDateAsString(ATTR_INSTANTIATEDATTRIBUTE_VALUE, d)
    }
    
    // enumeration?
    else if (typeName == "Enumeration") {
        writeXMLAttributeEnumeration(ATTR_INSTANTIATEDATTRIBUTE_VALUE, value)
    }
    
    // may be rich text
    else {
        writeXMLAttributeRichText(ATTR_INSTANTIATEDATTRIBUTE_VALUE, l, attributeName)
    }
    
    // end
    writeXMLEndTag();
}

void writeLinkInstantiatedAttributes(Link self) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES)) {
        // No need to write empty marker.
        // Link attributes cannot be requested on-demand.
        return;
    }
    
    writeXMLStartTag(START_TAG_ATTRIBUTEDELEMENT_INSTANTIATEDATTRIBUTES);
    string attributeName
    Module m = module(self);
    for attributeName in m do {
        writeInstantiatedAttribute(attributeName, self, m);
    }
    writeXMLEndTag();
}

void writeLink(Link self, Object source, Object target) {
    if (null self)
        return;
        
    // validate link
    if (source == null || isDeleted(source))
        return;
        
    if (target == null || isDeleted(target))
        return;
        
    if (OPTION_IGNORE_TABLE) {
        if (table(source) || row(source))
            return;
        if (table(target) || row(target))
            return;
    }
        
    // start
    writeXMLStartTag(START_TAG_LINK);
    writeXMLAttributeString(ATTR_LINK_SOURCE, getIDObject(source))
    writeXMLAttributeString(ATTR_EXTRA_LINK_SOURCE_URL, sodius_getURL(source));
    writeXMLAttributeString(ATTR_LINK_TARGET, getIDObject(target))
    writeXMLAttributeString(ATTR_EXTRA_LINK_TARGET_URL, sodius_getURL(target));
    
    // children
    // Note: always load the attributes, the link has no id and its children cannot be retrieved on-demand
    writeLinkInstantiatedAttributes(self);
    
    // end
    writeXMLEndTag();
}

void writeLinkSetLinks(Object self) {
    if (null self)
        return;
    if (! isFeatureRequested(FEATURE_LINKSET_LINKS))
        return;
        
    writeXMLStartTag(START_TAG_LINKSET_LINKS);
    
    // lookup source module
    Module sourceModule = getModule(self."source" "")
    if (! null sourceModule) {
        bool isSourceBaseline = isBaseline(sourceModule);

        // lookup target module
        string targetModuleFullName = self."target" "";
        Module targetModule = getModule(targetModuleFullName)
        if (! null targetModule) {
            bool isTargetBaseline = isBaseline(targetModule);
            
            // for each object in source module
            Link l;
            Object linkSource;
            string linkModuleName = fullName(module(self));
            for linkSource in all sourceModule do {
                
                // for each link starting from source object
                for l in all linkSource -> linkModuleName do {
                    checkCanceled();
                    
                    // In case of baseline, target(l) returns the object of the current module, not of the baseline!
                    // We need to lookup the object ourselves in the baseline.
                    if (isSourceBaseline || isTargetBaseline) {
                        
                        ModName_ targetName = target(l);
                        // ensure the link target is part of the current linkset target
                        if (! null targetName && fullName(targetName) == targetModuleFullName) {
                            Object linkTarget = getObject(targetAbsNo(l), targetModule);
                            if (! null linkTarget) {
                                writeLink(l, linkSource, linkTarget);
                            }
                        }
                    }
                    
                    else {
                        // ensure the link target is part of the current linkset target
                        Object linkTarget = target(l);
                        if (linkTarget != null && module(linkTarget) == targetModule) {
                            writeLink(l, linkSource, linkTarget);
                        }
                    }
                }
            }
        }
    }
    writeXMLEndTag();
}

void writeLinkSetChildren(Object self) {
    if (null self)
        return;
       
    writeLinkSetLinks(self);
}

void writeLinkSet(Object self, int depth) {
    if (null self)
        return;
        
    Item sourceItem = item(self."source" "");
    if (null sourceItem || getScopeMode(sourceItem) == SCOPE_MODE_EXCLUDE)
        return;
        
    Item targetItem = item(self."target" "");
    if (null targetItem || getScopeMode(targetItem) == SCOPE_MODE_EXCLUDE)
        return;
        
    // start
    writeXMLStartTag(START_TAG_LINKSET);
    writeXMLAttributeString(ATTR_XML_ID, getIDLinkSet(self))
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, fullName(sourceItem) " -> " fullName(targetItem))
    writeXMLAttributeString(ATTR_LINKSET_SOURCEMODULE, getIDItem(sourceItem))
    writeXMLAttributeString(ATTR_LINKSET_TARGETMODULE, getIDItem(targetItem))
    
    // children
    if (depth != DEPTH_ZERO) {
        writeLinkSetChildren(self);
    }
    
    // end
    writeXMLEndTag();
}

void writeLinkModuleLinkSetsEmpty(Module self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_LINKMODULE_LINKSETS))
        return;
    
    Object o
    for o in self do {
        return;
    }
    
    writeXMLStartTag(START_TAG_LINKMODULE_LINKSETS);
    writeXMLEndTag();
}

void writeLinkModuleLinkSets(Module self, Skip ids, int depth) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_LINKMODULE_LINKSETS)) {
        writeLinkModuleLinkSetsEmpty(self);
        return;
    }
    
    writeXMLStartTag(START_TAG_LINKMODULE_LINKSETS);
    Object o
    for o in self do {
        checkCanceled();
        writeLinkSet(o, depth); 
    }
    writeXMLEndTag();
}

bool isLinkModuleChildrenRequested(string id, Skip ids, int depth) {
    checkCanceled();
    
    if (! null ids) {
        int mode = 0;
        return find(ids, id, mode);
    }
    else if (depth != DEPTH_ZERO) {
        return isFeatureRequested(FEATURE_ATTRIBUTEDELEMENT_CREATEDBY) ||
        	isFeatureRequested(FEATURE_LINKMODULE_LINKSETS);
    }
    else {
        return false;
    }
}


void writeLinkModuleSystemAttributes(Module self) {
    if (null self) 
        return;
    
    writeModuleSystemAttributes(self);
    
    // mapping
    string mapping = self."Mapping";
    if (mapping == "One-to-Many") {
    	writeXMLAttributeString(ATTR_LINKMODULE_TYPE, "OneToMany");
    }
    else if (mapping == "One-to-One") {
    	writeXMLAttributeString(ATTR_LINKMODULE_TYPE, "OneToOne");
    }
    else if (mapping == "Many-to-One") {
    	writeXMLAttributeString(ATTR_LINKMODULE_TYPE, "ManyToOne");
    }
}

void writeLinkModuleChildren(Module self, Skip ids, int depth) {
    if (null self) 
        return;
    
    writeModuleSystemAttributes(self);
    writeModuleTypes(self);
    writeModuleAttributes(self);
    writeModuleInstantiatedAttributes(self);
    writeLinkModuleLinkSets(self, ids, depth);
}

bool isExcludedLinkModule(Item self, Skip ids, int depth) {
    if (isFilteredLinkModule(self))
        return true;
    else 
        return getScopeMode(self) == SCOPE_MODE_EXCLUDE;
}

void writeLinkModule(Item self, Skip ids, int depth) {
    if (null self) 
        return;
    if (isExcludedLinkModule(self, ids, depth))
        return;
        
    // start
    writeXMLStartTag(START_TAG_LINKMODULE);
    writeModuleItemSystemAttributes(self);
    
    // children
    if (isLinkModuleChildrenRequested(getIDItem(self), ids, depth)) {
        if (getScopeMode(self) == SCOPE_MODE_COMPLETE)
            writeLinkModuleChildren(getModule(self), null, DEPTH_LOCAL_INFINITE);
        else
            writeLinkModuleChildren(getModule(self), ids, decrementDepth(ids, depth));
    }
    else {
        // nothing, we don't want to load the module
    }
    
    // end
    writeXMLEndTag();
}

void writeFolderChildren(Folder, Skip, int);

bool isExcludedFolder(Folder self, Skip ids, int depth) {
    if (isFilteredFolder(self))
        return true;
    else 
        return getScopeMode(self) == SCOPE_MODE_EXCLUDE;
}

void writeBaselineSet(BaselineSet self) {
    if (null self) 
        return;
        
    writeXMLStartTag(START_TAG_BASELINESET);
    writeXMLAttributeInt(ATTR_BASELINEINFO_MAJOR, major(self));
    writeXMLAttributeInt(ATTR_BASELINEINFO_MINOR, minor(self));
    writeXMLAttributeString(ATTR_BASELINEINFO_SUFFIX, suffix(self));
    writeXMLAttributeString(ATTR_BASELINEINFO_DESCRIPTION, annotation(self));
    writeXMLAttributeString(ATTR_BASELINEINFO_CREATEDBY, user(self));
    writeXMLAttributeDate(ATTR_BASELINEINFO_CREATEDON, dateOf(self));
    writeXMLAttributeBool(ATTR_BASELINESET_OPEN, isOpen(self));
    
    // modules
    ModuleVersion moduleVersion;
    for moduleVersion in self do {
        writeXMLAttributeString(ATTR_BASELINESET_BASELINES, getIDBaseline(moduleVersion));
    }

    writeXMLEndTag();
}

void writeBaselineSetDefinition(BaselineSetDefinition self) {
    if (null self) 
        return;
        
    writeXMLStartTag(START_TAG_BASELINESETDEFINITION);
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, name(self));
    writeXMLAttributeString(ATTR_BASELINESETDEFINITION_DESCRIPTION, description(self));
    
    // baseline sets
    writeXMLStartTag(START_TAG_BASELINESETDEFINITION_BASELINESETS);
    BaselineSet set
    for set in self do {
        writeBaselineSet(set);
    }
    writeXMLEndTag();
    
    // modules
    ModName_ moduleName;
    for moduleName in self do {
        writeXMLAttributeString(ATTR_BASELINESETDEFINITION_MODULES, getIDModName(moduleName));
    }
    
    writeXMLEndTag();
}

void writeFolderBaselineSetDefinitionsEmpty(Folder self) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_FOLDER_BASELINESETDEFINITIONS))
        return;
        
    BaselineSetDefinition definition
    for definition in self do {
        return;
    }

    writeXMLStartTag(START_TAG_FOLDER_BASELINESETDEFINITIONS);
    writeXMLEndTag();
}

void writeFolderBaselineSetDefinitions(Folder self) {
    if (null self) 
        return;
    if (! isFeatureRequested(FEATURE_FOLDER_BASELINESETDEFINITIONS))
        return;
        
    writeXMLStartTag(START_TAG_FOLDER_BASELINESETDEFINITIONS);
    BaselineSetDefinition definition
    for definition in self do {
        writeBaselineSetDefinition(definition);
    }
    writeXMLEndTag();
}

void writeFolderOwnedItemsEmpty(Folder self, Skip ids, int depth) {
    if (null self) 
        return;
    if (! isEmptyFeatureRequested(FEATURE_AGGREGATEDELEMENT_OWNEDITEMS))
        return;
        
    Item child
    for child in self do {
        
        // folder
        if (type(child) == "Folder" || type(child) == "Project") {
            if (! isExcludedFolder(folder child, ids, depth)) {
                return;
            }
        }
        
        // formal module
        else if (type(child) == "Formal") {
            if (! isExcludedFormalModule(child, ids, depth)) {
                return;
            }
        }
        
        // link module
        else if (type(child) == "Link") {
            if (! isExcludedLinkModule(child, ids, depth)) {
                return;
            }
        }
    }

    writeXMLStartTag(START_TAG_AGGREGATEDELEMENT_OWNEDITEMS);
    writeXMLEndTag();
}

void writeFolderChildrenEmpty(Folder self, Skip ids, int depth) {
    writeFolderOwnedItemsEmpty(self, ids, depth);
    writeFolderBaselineSetDefinitionsEmpty(self);   
}

bool isFolderChildrenRequested(string id, Skip ids, int depth) {
    checkCanceled();
    
    if (! null ids) {
        int mode = 0;
        return find(ids, id, mode);
    }
    else if (depth != DEPTH_ZERO) {
        return isFeatureRequested(FEATURE_AGGREGATEDELEMENT_OWNEDITEMS);
    }
    else {
        return false;
    }
}

void writeFolder(Folder self, Skip ids, int depth) {
    if (null self) 
        return;
    if (isExcludedFolder(self, ids, depth))
        return;
    
    // start
    if (type(self) == "Project") {
        writeXMLStartTag(START_TAG_PROJECT);
    }
    else {
        writeXMLStartTag(START_TAG_FOLDER);
    }
    
    writeXMLAttributeString(ATTR_NAMEDELEMENT_NAME, name(self))
    writeXMLAttributeString(ATTR_SITUATEDELEMENT_URL, sodius_getURL(self))
    writeXMLAttributeString(ATTR_ITEM_DESCRIPTION, description(self))
    
    // children
    if (isFolderChildrenRequested(getIDFolder(self), ids, depth)) {
        if (getScopeMode(self) == SCOPE_MODE_COMPLETE) {
            writeFolderChildren(self, null, DEPTH_LOCAL_INFINITE);
        }
        else
            writeFolderChildren(self, ids, decrementDepth(ids, depth));
    }
    else {
        writeFolderChildrenEmpty(self, ids, depth);
    }
    
    // end
    writeXMLEndTag();
}

void writeDatabase() {
	Folder self = folder("/");
    if (null self) 
        return;
    
    writeXMLStartTag(START_TAG_AGGREGATEDELEMENT_OWNEDITEMS);
    writeXMLStartTag(START_TAG_FOLDER);
    writeXMLAttributeString(ATTR_SITUATEDELEMENT_URL, sodius_getURL(self));
    writeXMLAttributeString(ATTR_ITEM_DESCRIPTION, description(self));
    writeXMLAttributeString(ATTR_EXTRA_ENV_DOORSDATA, getenv("DOORSDATA"));
    writeXMLEndTag();
    writeXMLEndTag();
}

int getDefaultFormalModuleChildrenDepth(Skip ids, int depth) {
    if (! null ids)
        return depth;
    else if (depth == DEPTH_LOCAL_INFINITE)
        return depth;
    else
        return DEFAULT_FORMAL_MODULE_CHILDREN_DEPTH;
}

int getDefaultLinkModuleChildrenDepth(Skip ids, int depth) {
    if (! null ids)
        return depth;
    else if (depth == DEPTH_LOCAL_INFINITE)
        return depth;
    else
        return DEFAULT_LINK_MODULE_CHILDREN_DEPTH;
}

void writeFolderOwnedItems(Folder self, Skip ids, int depth) {
    if (null self) 
        return;
    if (isExcludedFolder(self, ids, depth))
        return;
    if (! isFeatureRequested(FEATURE_AGGREGATEDELEMENT_OWNEDITEMS)) {
        writeFolderOwnedItemsEmpty(self, ids, depth);
        return;
    }
        
    writeXMLStartTag(START_TAG_AGGREGATEDELEMENT_OWNEDITEMS);
    Item child
    for child in self do {
        checkCanceled();
        
        // folder
        if (type(child) == "Folder" || type(child) == "Project") {
            writeFolder(folder child, ids, depth);
        }
        
        // formal module
        else if (type(child) == "Formal") {
            writeFormalModule(child, ids, getDefaultFormalModuleChildrenDepth(ids, depth));
        }
        
        // link module
        else if (type(child) == "Link") {
            writeLinkModule(child, ids, getDefaultLinkModuleChildrenDepth(ids, depth));
        }
    }
    writeXMLEndTag();
}

void writeFolderChildren(Folder self, Skip ids, int depth) {
    if (null self) 
        return;
     
    writeFolderOwnedItems(self, ids, depth);   
    writeFolderBaselineSetDefinitions(self);   
}

void writeIDChildren(string id, Skip ids) {
    Item i;
    int objectAbsno = -1;
    if (decodeID(id, i, objectAbsno)) {
        string itemType = type(i);
        
        // object?
        if (objectAbsno >= 0 && itemType == "Formal") {
            Object o = getObject(objectAbsno, i)
            if (! null o) {
                writeObjectChildren(o, ids, 0);
            }  
        }
        
        // formal module?
        else if (itemType == "Formal") {
            writeFormalModuleChildren(getModule(i), ids, 0);
        }
        
        // link module?
        else if (itemType == "Link") {
            writeLinkModuleChildren(getModule(i), ids, 0);
        }
        
        // folder
        else if (itemType == "Folder" || itemType == "Project") {
            writeFolderChildren(folder(i), ids, 0);
        }
    }
} 

void addOnDemandParent(Folder f, Skip ids) {
    if (null f)
        return;
        
    // load this folder on-demand
    int mode = 0;
    string id = getIDFolder(f);
    if (! find(ids, id, mode)) {
        put(ids, id, SCOPE_MODE_ONDEMAND);
        put(SCOPE_MODES, id, SCOPE_MODE_ONDEMAND);
    }
    
    // load parent folder
    addOnDemandParent(getParentFolder(f), ids);
}

void writeCompleteItems(Skip completeIds) {
    Folder database = folder("/");
    
    // special case: load all database
    int mode = 0;
    if (find(completeIds, getIDFolder(database), mode)) {
        writeFolderChildren(database, null, DEPTH_LOCAL_INFINITE);
        return;
    }
    
    Skip ids = createString();
    
    for mode in completeIds do {
        string id = (string key completeIds);
        
        Item i = getItemById(id);
        if (! null i) {
            string itemId = getIDItem(i);
            put(ids, itemId, SCOPE_MODE_COMPLETE);
            put(SCOPE_MODES, itemId, SCOPE_MODE_COMPLETE);
            addOnDemandParent(getParentFolder(i), ids);
        }
    }
    
    writeFolderChildren(database, ids, 0);
    
    delete(ids);
}
