/*
 * @param xmlFilename - output XML filename
 */
 
Stream xmlFile
Buffer xmlBuffer = create
bool xmlStartTagClosed = true;

const string EVENT_START_TAG = "0|"
const string EVENT_START_TAG_END = "1|"
const string EVENT_END_TAG = "2|"
const string EVENT_END_EMPTY_TAG = "3|"
const string EVENT_ATTRIBUTE_STRING = "4|"
const string EVENT_ATTRIBUTE_NON_STRING = "5|"

const string SEPARATOR = "|"

const string ATTR_XML_ID = EVENT_ATTRIBUTE_STRING "1000|"
const string ATTR_EXTRA_LINK_SOURCE_URL = EVENT_ATTRIBUTE_STRING "1001|"
const string ATTR_EXTRA_LINK_TARGET_URL = EVENT_ATTRIBUTE_STRING "1002|"
const string ATTR_EXTRA_ENV_DOORSDATA = EVENT_ATTRIBUTE_STRING "1003|"

int imageCount = 0;

void flushXMLBuffer() {
	if (length(xmlBuffer) == 0)
		return;
		
	xmlFile << xmlBuffer
	setempty xmlBuffer;
}

void requestFlushXMLBuffer() {
	if (length(xmlBuffer) > 512) {
		flushXMLBuffer()
	}
}

void writeXMLStartTag(string tagId) {
	xmlBuffer += tagId
	xmlStartTagClosed = false;
}

void writeXMLStartTagEnd() {
	xmlBuffer += EVENT_START_TAG_END
	xmlStartTagClosed = true;
	requestFlushXMLBuffer();
}

void writeXMLEndTagEmpty() {
    xmlBuffer += EVENT_END_EMPTY_TAG
    xmlStartTagClosed = true;
    requestFlushXMLBuffer();
}

void writeXMLEndTag() {
    if (! xmlStartTagClosed) {
        writeXMLEndTagEmpty();
    }
    else {
    	xmlBuffer += EVENT_END_TAG
    	requestFlushXMLBuffer();
    }
}

string createImageFilename(int imageFormat) {
	imageCount++;
	return IMAGES_FOLDER imageCount ".png"
}

Regexp richTextDetectionRegex = regexp "\\\\(b|i|ul|tab|sub|super|strike|object|wmetafile|fcharset. Symbol)";

bool isReallyRichText(Buffer b) {
    int startIndex = 0; 
    return search(richTextDetectionRegex, b, startIndex); 
}

// If DOORS client is >= 9.3, use a DXL function that returns the OLE content in a buffer
// rather than in a string to avoid memory leaks. 
string getAddOleSourceCode() {
	if (isDoorsClientCompliantWith('9', '3')) {
		return "
		Buffer addOleBuffer = create();
		void addOLEContent(Buffer bufferXHTML, RichText rt) { 
		   combine(bufferXHTML, oleRtf(rt.getEmbeddedOle, addOleBuffer), 0);
		}";
	}
	else {
		return "
		void addOLEContent(Buffer bufferXHTML, RichText rt) { 
		   bufferXHTML += oleRtf(rt.getEmbeddedOle);
		}";
	}
}

Buffer bufferRTF    = create(); 
bool isFormattedRef = false;   
Buffer bufferXHTML = create();
Buffer bufferOLE = create();
Buffer bufferValue = create(2048);

string sToXHTMLCode = "
	pragma runLim, 0 
	pragma encoding, \"UTF-8\"
	" getAddOleSourceCode() "
    char symbolTable[] = {' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ','!','∀','#','∃','%','&','∍','(',')','∗','+',',','−','.','/','0','1','2','3','4','5','6','7','8','9',':',';','<','=','>','?','≅','Α','Β','Χ','Δ','Ε','Φ','Γ','Η','Ι','ϑ','Κ','Λ','Μ','Ν','Ο','Π','Θ','Ρ','Σ','Τ','Υ','ς','Ω','Ξ','Ψ','Ζ','[','∴',']','⊥','_','`','α','β','χ','δ','ε','φ','γ','η','ι','ϕ','κ','λ','μ','ν','ο','π','θ','ρ','σ','τ','υ','ϖ','ω','ξ','ψ','ζ','{','|','}','∼','','€','ﾁ','‚','ƒ','„','…','†','‡','ˆ','‰','Š','‹','Œ','ﾍ','Ž','ﾏ','ﾐ','‘','’','“','”','•','–','—','˜','™','š','›','œ','ﾝ','ž','Ÿ',' ','ϒ','′','≤','⁄','∞','ƒ','♣','♦','♥','♠','↔','←','↑','→','↓','°','±','″','≥','×','∝','∂','∙','÷','≠','≡','≈','…','⏐','⎯','↵','ℵ','ℑ','ℜ','℘','⊗','⊕','∅','∩','∪','⊃','⊇','⊄','⊂','⊆','∈','∉','∠','∇','®','©','™','∏','√','⋅','¬','∧','∨','⇔','⇐','⇑','⇒','⇓','◊','〈','®','©','™','∑','⎛','⎜','⎝','⎡','⎢','⎣','⎧','⎨','⎩','⎪','€','〉','∫','⌠','⎮','⌡','⎞','⎟','⎠','⎤','⎥','⎦','⎫','⎬','⎭','ÿ'};
    void appendSymbols(Buffer buffer, string text) {
        int length = length(text);
        if (length == 0)
            return;
           
        int i = 0; 
        for (i = 0; i < length; i++) {
            char c = symbolTable[intOf text[i]];
            if (c == '&') buffer += \"&amp;\"
            else if (c == '<')  buffer += \"&lt;\"
            else if (c == '>')  buffer += \"&gt;\"
            else buffer += c
        }
    }
    
    void appendXHTMLText(Buffer buffer, string text) {
        int length = length(text);
        if (length == 0)
            return;
           
        int i = 0; 
        for (i = 0; i < length; i++) {
            char c = text[i]
            if (c == '&') buffer += \"&amp;\"
            else if (c == '\\'') buffer += \"&apos;\"
            else if (c == '\\\"') buffer += \"&quot;\"
            else if (c == '<')  buffer += \"&lt;\"
            else if (c == '>')  buffer += \"&gt;\"
            else if (c == '\\n') buffer += \"&#10;\"
            else if (c == '\\r') buffer += \"&#13;\"
            else if (c == '\\t') buffer += \"&#9;\"
            else buffer += c
        }
    }
    
    Buffer &bufferRTF = addr_(" to32BitLiteral(addr_ (&bufferRTF)) ") 
    bool &isFormatted = addr_(" to32BitLiteral(addr_ (&isFormattedRef)) ") 
    Buffer &bufferXHTML = addr_(" to32BitLiteral(addr_ (&bufferXHTML)) ") 
    Buffer &bufferOLE = addr_(" to32BitLiteral(addr_ (&bufferOLE)) ") 
    bool &OPTION_IGNORE_OLE_PREVIEW_IMAGES = addr_(" to32BitLiteral(addr_ (&OPTION_IGNORE_OLE_PREVIEW_IMAGES)) ") 
    string &IMAGES_FOLDER = addr_(" to32BitLiteral(addr_ (&IMAGES_FOLDER)) ") 
    int &imageCount = addr_(" to32BitLiteral(addr_ (&imageCount)) ") 
    
    RichTextParagraph rp
    RichText rt
    string closingBulletTag = null;
    int listBulletStyle = 0;
    
    // for each paragraph
    for rp in (tempStringOf bufferRTF) do {
        // bullet paragraph
        if (rp.isBullet) {
            
            // li style
            string itemStyle = null;
            if (rp.bulletStyle == 1) {
                itemStyle = \"disc\";
            }
            else if (rp.bulletStyle == 2) {
                itemStyle = \"decimal\";
            }
            else if (rp.bulletStyle == 3) {
                itemStyle = \"lower-alpha\";
            }
            else if (rp.bulletStyle == 4) {
                itemStyle = \"upper-alpha\";
            }
            else if (rp.bulletStyle == 5) {
                itemStyle = \"lower-roman\";
            }
            else if (rp.bulletStyle == 6) {
                itemStyle = \"upper-roman\";
            }

            // first item of the list?
            if (null closingBulletTag) {
                listBulletStyle = rp.bulletStyle;
                if (rp.bulletStyle == 1) {
                    bufferXHTML += \"<ul>\";
                    closingBulletTag = \"</ul>\";
                }
                else if (rp.bulletStyle == 2) {
                    bufferXHTML += \"<ol>\";
                    closingBulletTag = \"</ol>\";
                }
                else {
                    bufferXHTML += \"<ol style=\\\"list-style-type:\" itemStyle \"\\\">\";
                    closingBulletTag = \"</ol>\";
                }
                isFormatted = true;
            }
            
            int margin = 30 * ((rp.indentLevel / 360) - 1);
            bufferXHTML += \"<li\";
            if (margin > 0 || rp.bulletStyle != listBulletStyle) {
                bufferXHTML += \" style=\\\"\";
                if (margin > 0) {
                    bufferXHTML += \"margin-left: \" margin \"px;\";
                }
                if (rp.bulletStyle != listBulletStyle) {
                    bufferXHTML += \"list-style-type:\" itemStyle;
                }
                bufferXHTML += \"\\\"\";
            }
            bufferXHTML += \">\";
        }
        else {
            if (! null closingBulletTag) {
                bufferXHTML += closingBulletTag
                closingBulletTag = null;
                listBulletStyle = 0;
            }
            bufferXHTML += \"<p>\"
        }
        
        // for each chunk
        for rt in rp do
        {            
            // ole?
            if (rt.isOle) {
                bufferXHTML += \"<object data=\\\"data:\";
                string oleFilename = tempFileName();
                bufferXHTML += oleFilename;
                bufferXHTML +=  \"\\\"\";
                if (! OPTION_IGNORE_OLE_PREVIEW_IMAGES) {
                	imageCount++
                	string imageFilename = IMAGES_FOLDER imageCount \".png\";
                	if (null exportPicture(rt.getEmbeddedOle, imageFilename, formatPNG)) {
	                	bufferXHTML += \" preview=\\\"\" imageFilename \"\\\"\";
                	}
                }
                bufferXHTML +=  \"/>\";
                isFormatted = true;
                
                Stream oleStream = write binary oleFilename;
                setempty(bufferOLE);
                addOLEContent(bufferOLE, rt);
                oleStream << bufferOLE;
                setempty(bufferOLE);
                close(oleStream);
            }
                
            // open formatting tags (reverse order of close)
            if (rt.strikethru) {
                bufferXHTML += \"<del>\"
                isFormatted = true;
            }
            if (rt.underline) {
                bufferXHTML += \"<u>\"
                isFormatted = true;
            }
            if (rt.italic) {
                bufferXHTML += \"<i>\"
                isFormatted = true;
            }
            if (rt.bold) {
                bufferXHTML += \"<b>\"
                isFormatted = true;
            }
            if (rt.subscript) {
                bufferXHTML += \"<sub>\"
                isFormatted = true;
            }
            if (rt.superscript) {
                bufferXHTML += \"<sup>\"
                isFormatted = true;
            }
            
            // text
            // Note: in case of OLE, the text always contain one space, which does not exist in DOORS
            if (! rt.isOle) {
                if (rt.charset == charsetSymbol) {
                    appendSymbols(bufferXHTML, rt.text);
                    isFormatted = true;
                }
                else {
                    appendXHTMLText(bufferXHTML, rt.text);
                }
            }
            
            // close formatting tags
            if (rt.superscript) {
                bufferXHTML += \"</sup>\"
            }
            if (rt.subscript) {
                bufferXHTML += \"</sub>\"
            }
            if (rt.bold) {
                bufferXHTML += \"</b>\"
            }
            if (rt.italic) {
                bufferXHTML += \"</i>\"
            }
            if (rt.underline) {
                bufferXHTML += \"</u>\"
            }
            if (rt.strikethru) {
                bufferXHTML += \"</del>\"
            }
        }
        
        // end paragraph
        if (rp.isBullet) {
            bufferXHTML += \"</li>\"
        }
        else {
            bufferXHTML += \"</p>\"
        }
    }
    
    if (isFormatted) {
        // close bullet
        if (! null closingBulletTag) {
            bufferXHTML += closingBulletTag;
        }
    }    
";

// Result will be returned in global buffer bufferXHTML.
// Use eval_() function to avoid memory leaks.
void toXHTML(Buffer value) {
    setempty bufferXHTML
        
    // RTF?
    if (OPTION_RICHTEXT_RTF) {
    	isFormattedRef = sodius_rtf_preprocess(value);
	    if (! isFormattedRef && ! isReallyRichText(value)) {
	    	return;
	    }

        bufferXHTML += value;
    }
    else {
        // XHTML
        isFormattedRef = sodius_rtf_preprocess(value);
	    if (! isFormattedRef && ! isReallyRichText(value)) {
	    	return;
	    }
    
        bufferXHTML = "<html><body>";
    
        bufferRTF = value; 
        eval_ sToXHTMLCode
        
        if (isFormattedRef) {
            bufferXHTML += "</body></html>";
        } else {
            setempty bufferXHTML;
        }
    }
}

void writeXMLAttributeString(string attrId, string value) {
    if (value == null || value == "")
        return;
        
    xmlBuffer += attrId
    int l = length(value)
    xmlBuffer += l SEPARATOR 
    xmlBuffer += value
    xmlBuffer += SEPARATOR
    flushXMLBuffer();
}

void writeXMLAttributeStringEmpty(string attrId) {
    xmlBuffer += attrId
    xmlBuffer += 0 SEPARATOR 
    xmlBuffer += ""
    xmlBuffer += SEPARATOR
    flushXMLBuffer();
}

string formatEnumeration(string value) {
    Buffer buffer = create
    int length = length(value);
    int i = 0;
    for (i = 0; i < length; i++) {
        char c = value[i]
        if (c == '\n') 
            buffer += OPTION_ENUMERATION_LITERAL_SEPARATOR;
        else 
            buffer += c
    }
    
    value = stringOf(buffer);
    delete buffer;
    return value;
}

void writeXMLAttributeEnumeration(string attrId, string value) {
    if (value == null || value == "")
        return;
        
    writeXMLAttributeString(attrId, formatEnumeration(value));
}

void writeXMLAttributeRichText(string attrId, string value, Buffer richTextValue) {
    xmlBuffer += attrId
    
    toXHTML(richTextValue);
    int l = length(bufferXHTML)
    
    if (l != 0) {
        xmlBuffer += "R" l SEPARATOR; // 'R' for rich text
        combine(xmlBuffer, bufferXHTML, 0);
    } 
    else {
        int l = length(value);
        xmlBuffer += l SEPARATOR; 
        xmlBuffer += value;
    }
    
    xmlBuffer += SEPARATOR
    flushXMLBuffer();
}

void writeXMLAttributeRichText(string attrId, Object o, string attributeName) {
    string value = o.attributeName ""
    if (value == null || length(value) == 0)
        return;
        
    if (! OPTION_IGNORE_RICHTEXT) {
    	if (OPTION_IGNORE_OLE_OBJECTS) {
    	    bufferValue = richText(o.attributeName);
	        writeXMLAttributeRichText(attrId, value, bufferValue);
    	}
    	else {
    	    bufferValue = richTextWithOle(o.attributeName);
	        writeXMLAttributeRichText(attrId, value, bufferValue);
    	}
    }
    else {
        writeXMLAttributeString(attrId, value);
    }
}

void writeXMLAttributeRichText(string attrId, Module m, string attributeName) {
    string value = m.attributeName ""
    if (value == null || length(value) == 0)
        return;
        
    if (! OPTION_IGNORE_RICHTEXT) {
    	bufferValue = richText(m.attributeName);
    	writeXMLAttributeRichText(attrId, value, bufferValue);
    }
    else {
        writeXMLAttributeString(attrId, value);
    }
}

void writeXMLAttributeRichText(string attrId, Link l, string attributeName) {
    string value = l.attributeName ""
    if (value == null || length(value) == 0)
        return;
        
    if (! OPTION_IGNORE_RICHTEXT) {
    	bufferValue = richText(l.attributeName);
        writeXMLAttributeRichText(attrId, value, bufferValue);
    }
    else {
        writeXMLAttributeString(attrId, value);
    }
}

void writeXMLAttributeInt(string attrId, int value) {
    xmlBuffer += attrId 
    xmlBuffer += value SEPARATOR;
}

void writeXMLAttributeBool(string attrId, bool value) {
    xmlBuffer += attrId 
    xmlBuffer += value SEPARATOR;
}

void writeXMLAttributeDate(string attrId, Date value) {
    if (null value)
        return;
   
    string s = stringOf(value, DATE_LOCALITY, DATE_FORMAT)
    xmlBuffer += attrId 
    xmlBuffer += s
    xmlBuffer += SEPARATOR
}

void writeXMLAttributeDateAsString(string attrId, Date value) {
    if (null value)
        return;
   
    writeXMLAttributeString(attrId, stringOf(value, DATE_LOCALITY, DATE_FORMAT));
}

void writeXMLStartDocument() {
	xmlFile = write(xmlFilename, DOORS_CODEPAGE)
}

void writeXMLEndDocument() {
	flushXMLBuffer();
	close(xmlFile)
	delete xmlBuffer
}
