pragma runLim, 0 
pragma encoding, "UTF-8"

// Common functions -----------------------------------------------------
bool connector_checkFileExists(string sFilename) {
    Stat s = create(sFilename);
    if (null s) {
    	return false;
    }
    else { 
    	delete s; 
    	return true;
    }
}

Stream connector_createStream(string sFileName) {
	Stream st = null;
	int iRetryCount = 20;
	while (true) { 
		noError();
		st = write sFileName;
    	if (null lastError() || iRetryCount -- == 0) { 
    		break;
    	}
    	sleep_ 20;
	}
	return st;
}     

void connector_deleteFile(string sFileName) {
	int retryCount = 20; 
    while (retryCount > 0) { 
    	deleteFile sFileName;
		if (!connector_checkFileExists(sFileName)) { 
			break;
		}
		retryCount--; 
		sleep_ 20;
    }
}

bool connector_createFlagFile(string sFileName) {
	Stream st = connector_createStream(sFileName);      
    if (null st) { 
    	return false;
    }
    close st; 
    return true;
}

// Request -----------------------------------------------------
bool connector_isAlive() {
	return connector_checkFileExists(connector_sFileFlagAlive);
}

bool connector_isRequestAvailable() {
	return connector_checkFileExists(connector_sFileFlagRequest);
}

void connector_deleteRequestFlag() {
	connector_deleteFile(connector_sFileFlagRequest);
}

string connector_readRequestFilename() {
	string filename = null;
	Stream st = read(connector_sFileDataRequest, CP_UTF8);
	if (! null st) {
		st >> filename;
		close(st);
	}
	return filename;
}

// Response -----------------------------------------------------
void connector_notifyResponse() {
	connector_createFlagFile(connector_sFileFlagResponse);
}

void connector_execute(string filename) {
	eval_("#include <" filename ">")
}

// Main -----------------------------------------------------
while(connector_isAlive()) {
	
	// request?
	if (connector_isRequestAvailable()) {
		
		// read request filename
		string filename = connector_readRequestFilename();
		if (null filename || length(filename) == 0) {
			print "Internal Error: Failed to read the request filename";
			halt;
		}
		else {
			// execute request
			connector_deleteRequestFlag();
			connector_execute(filename);
			
			// mark as completed
			connector_notifyResponse();
		}
	}
	
	// wait a bit
	else {
		sleep_ 20;
	}
}

