bool existModule(string fullName) {
	Item i = item(fullName);
	if (null i) {
		return false;
	}
	else {
		return type(i) == "Formal";
	}
}

Module createModule(string fullName, string description, string prefix, bool display) {
    Module m = edit(fullName, display, true /* standard view */);
    if (null m) {
        m = read(fullName, false);
        if (null m) {
            m = create(fullName, description, prefix, 1 /* absno */, false /* no display */);
        }
        if (!null m) {
            close(m);
            m = edit(fullName, display, true /* standard view */);
        }
    }
    return m;
}

void flattenObjectHierarchy(Module m) {
    Object referee = first m;
    while ((!null referee) and (level(referee) != 1)) {
        referee = next referee;
    }
    if (!null referee) {
    	Object obj;
        int size = 0;
        for obj in m do {
            size++
        }
        Array objs = create(size, 1)
        int i = 0;
        for obj in all(m) do {
            put(objs, obj, i, 0);
            i++ ;
        }

        for (i = 0; i < size; i++) {
            obj = (Object get(objs,i,0));
            if (level(obj) != 1) {
                if (canModify(obj) and !cell(obj) and !row(obj)) {
                    move(obj, after referee);
                }
            }
        }
        
        delete(objs);
    }
}

void deleteAllObjects(Module m) {
    noError(); // we might have ignorable errors when an object contains a picture
    Object obj;
    for obj in top(m) do {
        softDelete(obj);
    }
    lastError();
}

void writeAbsoluteNumbersFile(Object objects[], string filename) {
    Buffer buffer = create();
    int i;
    int l = sizeof(objects);
    for (i = 0; i < l; i++) {
    	if (null objects[i]) {
        	buffer += "\n";
    	}
    	else {
        	buffer += objects[i]."Absolute Number" "\n";
    	}
    }
    
    Stream file = write filename
    file << utf8(stringOf buffer)
    close(file)
    delete(buffer);
}

void moveTopObject(Object o, Object sibling) {
    if (canModify(o) && !cell(o) && !cell(sibling)) {
        if (o != sibling) {
            move(o, after sibling);
        }
    }
}

void moveNestedObject(Object o, Object p) {
    if (canModify(o)) {
        
        if (cell(o))
            o = parent(o);
        if (row(o))
            o = parent(o);
            
        if (o != p) {
            move(o, below p)
        }
    }
}

Object replaceImage(Object o, string imageFile) {
	int pictureFormat = getPictFormat(o);
	if (pictureFormat > 0) {
		// replace picture
		insertPictureFile(imageFile, formatBMP, o);
		return o;
	}
	else {
		// not a picture, need to insert a new object
		insertPictureFileAfter(imageFile, formatBMP, o);
		return next(o);
	}
}

Object getLastRootObject(Module m) {
	
	// last of module (in depth first order)
	Object o = last(m);
	
	while (! null(o)) {
		
		// has parent -> goto parent
		Object parentObject = parent(o);
		if (! null parentObject) {
			o = parentObject;
		}
		
		// deleted -> goto previous
		else if (isDeleted(o)) {
			o = previous(o);
		}
		
		// root and not deleted-> return
		else {
			return o;
		}
	}
	
	// no root
	return null;
}

Object createObject(Module m) {
    if (canCreate(m)) {
    	Object lastRootObject = getLastRootObject(m);
    	if (null lastRootObject) {
            return create(m);
    	}
    	else {
        	return create(after lastRootObject);
    	}
    }
    else {
        return null;
    }
}

Object createObject(Module m, int absno) {
    Object o = object(absno, m);
    if (null o) {
        o = createObject(m);
    }
    return o;
}

Object createObject(Module m, string imageFile) {
    if (canCreate(m)) {
    	Object lastRootObject = getLastRootObject(m);
    	if (null lastRootObject) {
            insertPictureFileAfter(imageFile, formatBMP, null);
            return getLastRootObject(m);
    	}
    	else {
            insertPictureFileAfter(imageFile, formatBMP, lastRootObject);
            return next(lastRootObject);
    	}	
    }
    return null;
}

Object createObject(Module m, int absno, string imageFile) {
    Object o = object(absno, m);
    if (null o) {
        o = createObject(m, imageFile);
    }
    else {
    	o = replaceImage(o, imageFile);
    }
    return o;
}

Object createObject(Object parentObject) {
    return create(last(below(parentObject)));
}

Object createObject(Object parentObject, int absno) {
    Object o = object(absno, module(parentObject));
    if (null o) {
        o = createObject(parentObject);
    }
    return o;
}

Object createObject(Object parentObject, string imageFile) {
    insertPictureFileAfter(imageFile, formatBMP, parentObject);
    return next(parentObject);
}

Object createObject(Object parentObject, int absno, string imageFile) {
    Object o = object(absno, module(parentObject));
    if (null o) {
        o = createObject(parentObject, imageFile);
    }
    else {
    	o = replaceImage(o, imageFile);
    }
    return o;
}

void formatTable(Object tableObject) {
	setAllCellsBorder(tableObject, left, solidBorder);
	setAllCellsBorder(tableObject, right, solidBorder);
	setAllCellsBorder(tableObject, top, solidBorder);
	setAllCellsBorder(tableObject, bottom, solidBorder);
}

Object createTable(Module m, int rows, int cols) {
	Object tableObject = table(m, rows, cols);
	formatTable(tableObject);
	return tableObject;
}

Object createTable(Object parentObject, int rows, int cols) {
	Object tableObject = table(last(below(parentObject)), rows, cols);
	formatTable(tableObject);
	return tableObject;
}

Object getRow(Object tableObject, int rowIndex) {
	int currentRow = 0;
	Object rowObject = null;
	
	for rowObject in table(tableObject) do {
		if (currentRow == rowIndex) {
			return rowObject;
		}
		else {
			currentRow++;
		}
	} 
	
	return null;
}

Object getCell(Object rowObject, int columnIndex) {
	int currentCell = 0;
	Object cellObject = null;
	
	for cellObject in row(rowObject) do {
		if (currentCell == columnIndex) {
			return cellObject;
		}
		else {
			currentCell++;
		}
	} 
	
	return null;
}

Object getCell(Object rowObject,int columnIndex, int absno) {
	Object o = object(absno, module(rowObject));
	if (null o) {
		o = getCell(rowObject, columnIndex);
	}
	return o;
}

Object getCell(Object rowObject, int columnIndex, string imageFile) {
	return getCell(rowObject, columnIndex);
}

Object getCell(Object rowObject, int columnIndex, int absno, string imageFile) {
	return getCell(rowObject, columnIndex, absno);
}

void setColumnWidths(Object tableObject, int columnWidths[]) {
	Object rowObject = null;
	for rowObject in table(tableObject) do {
		
		int i = 0;
		for (i = 0; i < sizeof columnWidths; i++) {
			setColumnWidth(getCell(rowObject, i), columnWidths[i]);
		}
		
		return;
	}
}

bool canModifyObjectText(Object o) {
    
    // bug id=3890: Do not overwrite an object having an OLE object 
    if (oleIsObject(o)) {
        return false;
    }
    
    return true;
}

void setObjectText(Object o, string value) {
    if (canModifyObjectText(o)) {
        setStringAttr(o, "Object Text", value);
    }
}

ExternalLink createExternalLink(Object o, string description, string name, ExternalLinkDirection extLinkDir, ExternalLinkBehaviour extLinkBeh, string body) {
    ExternalLink el;
    create(o, description, name, extLinkDir, extLinkBeh, body, el);
    return el;
}

void deleteColumns(Module m) {
    
    // count the number of columns
    int n = 0; 
    Column c;
    for c in m do {
        n++;
    }
    
    // delete each column
    if (n > 0) { 
        int i;
        for i in 1 : n do {
            delete column 0;
        } 
    } 
}

void importImage(Object o, string filename) {
	insertPictureFileAfter(filename, formatBMP , null);
}

void setObjectTextRTF(Object o, string filepath) {
	if (null o) {
		return;
	}
	
	// already has an object text
	string currentText = richText(o."Object Text");
	if (length(currentText) > 0) {
	
		// detect end of rtf
		int indexEndRtf = length(currentText) - 1;
		while (indexEndRtf > 0 && currentText[indexEndRtf] != '}') {
			indexEndRtf--;
		}
		if (indexEndRtf > 0) {
			
			// append RTF file to current text
			Buffer b = create();
			b += currentText[0:indexEndRtf]
			b += readFile(filepath);
			b += "}\n";
			
			// change Object Text
			o."Object Text" = richText(stringOf(b));
			delete(b);
			
			// we're done
			return;
		}
	}
	else {
		o."Object Text" = richText(readFile(filepath));
	}
}

void setObjectTextOLE(Object o, string filepath) {
	if (null o) {
		return;
	}
	
	// need to add an extra space (eaten by oleInsert())
	string s = o."Object Text";
	if (length(s) > 0) {
		o."Object Text" = "" o."Object Text" " ";
	}
	
	// insert OLE
	oleInsert(o, filepath);
}
