
Date parseDate(string value) {
    if (null value || length(value) == 0)
        return null;
    else
        return date(value, DATE_LOCALITY, DATE_FORMAT);
}

string formatDate(string value) {
    if (null value || length(value) == 0) {
        return "";
    }
    else {
        Date d = parseDate(value);
        return d "";
    }
}

void deleteAttributesAndTypes(Module m) {
    
    AttrDef at;
    for at in m do {
        if (canDelete(m)) {
            delete(at)
        }
    }
    
    string errMess;
    AttrType t;
    for t in m do {
        if (!t.system) {
            delete(t, errMess);
        }
    }
}

bool hasAttributes(Module m) {
    AttrDef a;
    for a in m do {
        if (! a.system) {
            return true;
        }
    }
    
    return false;
}

int getAttributesSize(Module m) {
    int result = 0;
    AttrDef a;
    for a in m do {
        if (! a.system) {
            result++;
        }
    }
    return result;
}

void copyAttributeNames(Module m, string result[]) {
    if (sizeof(result) == 0)
        return;
        
    AttrDef a;
    int i = 0;
    for a in m do {
        if (! a.system) {
            result[i] = a.name;
            i++;
        }
    }
}

bool isRTFText(string value) {
    if (null value)
        return false;
        
    if (length(value) > 4) {
        if (value[0] == '{' && value[1] == '\\' && value[2] == 'r' && value[3] == 't' && value[4] == 'f') {
            return true;
        }
    }
        
    return false;
}

void setStringAttr(Object o, string name, string value) {
    if (isRTFText(value))
        o.name = richText(value);
    else
        o.name = value;
}

void setStringAttr(Module o, string name, string value) {
    if (isRTFText(value))
        o.name = richText(value);
    else
        o.name = value;
}

void setStringAttr(Link o, string name, string value) {
    if (isRTFText(value))
        o.name = richText(value);
    else
        o.name = value;
}

void updateAttribute(AttrDef at, bool pModule, bool pObject, bool pChangeBars, bool pDate, string pDxl, bool pHistory, bool pInherit, bool pMulti, string pDefault) {
    modify(at, setModule, pModule);
    modify(at, setObject, pObject);
    modify(at, setBars, pChangeBars);
    modify(at, setDates, pDate);
    modify(at, setHistory, pHistory);
    modify(at, setInherit, pInherit);
    modify(at, setMulti, pMulti);
    
    if (! null pDxl && length(pDxl) != 0) {
        modify(at, setDXL, pDxl);
    }
    
    if (! null pDefault && length(pDefault) != 0) {
        modify(at, setDefault, pDefault);
    }
}

AttrDef createAttribute(Module m, string pName, string pTypeName, bool pModule, bool pObject, bool pChangeBars, bool pDate, string pDxl, bool pHistory, bool pInherit, bool pMulti, string pDefault) {
    AttrDef at = find(m, pName); 
    if (! null at) {
        if (at.typeName != pTypeName) {
            if (canDelete(m) && canCreate(m)) {
                delete(at);
                at = create type pTypeName attribute pName;
            }
            else if (canModify(m)) {
                modify(at, type pTypeName attribute pName);
            }
        }
    }
    else if (canCreate(m)) {
        at = create type pTypeName attribute pName;
    }
    
    if (! null at) {
        updateAttribute(at, pModule, pObject, pChangeBars, pDate, pDxl, pHistory, pInherit, pMulti, pDefault)
    }
    
    return at;
}

bool contains(string array[], string value) {
    if (null array) {
        return false;
    }
        
    int i;
    int l = sizeof(array);
    for (i = 0; i < l; i++) {
        if (array[i] == value)
            return true;
    }
    
    return false;
}

void deleteObsoleteAttributes(Module m, string validNames[]) {
    AttrDef a;
    for a in m do {
        if (! contains(validNames, a.name)) {
            if (canDelete(m, a.name)) {
                delete(a);
            }
        }
    }
}

void writeAbsoluteNumbersFile(Object objects[], string filename) {
    Buffer buffer = create();
    int i;
    int l = sizeof(objects);
    for (i = 0; i < l; i++) {
        buffer += objects[i]."Absolute Number" "\n";
    }
    
    Stream file = write filename
    file << utf8(stringOf buffer)
    close(file)
    delete(buffer);
}

void deleteAttributes(Module m, AttrType t) {
    AttrDef a;
    for a in m do {
        if (a.type == t) {
            delete(a);
        }
    }
}

bool hasType(Module m, string name) {
    AttrType t = find(m, name);
    return ! null t;
}

bool hasType(Module m, string name, AttrBaseType baseType) {
    AttrType t = find(m, name);
    if (null t) {
        return false;
    }
    else {
        if (t.type == baseType) {
            return true;
        }
        else {
            deleteAttributes(m, t);
            
            string errMess;
            delete(t, errMess);
            
            return hasType(m, name);
        }
    }
}

AttrType createType(string name, AttrBaseType baseType) {
    string errmsg = null;
    AttrType t = create(name, baseType, errmsg);
    if (! null errmsg) {
        print errmsg;
    }
    return t;
}

AttrType createEnumeratedType(string name, string codes[], int values[]) {
    string errmsg = null;
    AttrType t = create(name, codes, values, errmsg);
    if (! null errmsg) {
        print errmsg;
    }
    return t;
}

AttrType createType(string name, AttrBaseType baseType, int min, int max) {
    AttrType t = createType(name, baseType);
    if (! null t) {
        setMinValue(t, min, true);
        setMaxValue(t, max, true);
    }
    return t;
}

AttrType createType(string name, AttrBaseType baseType, real min, real max) {
    AttrType t = createType(name, baseType);
    if (! null t) {
        setMinValue(t, min, true);
        setMaxValue(t, max, true);
    }
    return t;
}

AttrType createType(string name, AttrBaseType baseType, Date min, Date max) {
    AttrType t = createType(name, baseType);
    if (! null t) {
        if (! null min) setMinValue(t, min, true);
        if (! null max) setMaxValue(t, max, true);
    }
    return t;
}

void saveAndClose(Module m) {
    save(m);
    close(m);
}

void saveAndClose(string fullName) {
    Item i = item(fullName);
    if (! null i) {
        Module m = module(i);
        if (! null m) {
            saveAndClose(m);
        }
    }
}
