pragma encoding, "UTF-8"

const int MODULE_MODE_READ = 0;
const int MODULE_MODE_EDIT = 1;
const int MODULE_MODE_SHARE = 2;

Skip openedModules = null
Skip lockedObjects = null
string openModulesFilename = null

void sodius_setResult(string message) {
    
    // verify a message is not already logged
    if (null resultFilename) {
        return;
    }
    
    Stream file = write(resultFilename, CP_UTF8);
    file << message;
    close(file);
    
    resultFilename = null;
}

void sodius_setResultSuccess() {
    sodius_setResult("SUCCESS");
}

bool sodius_isTerminated() {
    return null resultFilename;
}

void sodius_setError(string message) {
    if (null message || length(message) == 0) {
        sodius_setResult("Empty DOORS error message");
    }
    else {
        sodius_setResult(message);
    }
}

string sodius_getMessageOrLastError(string message) {
    string l = lastError();
    if (null message) {
        return message;
    }
    else {
        return l;
    }
}

Module getOpenModule(string moduleFullName) {
	Item item = item(moduleFullName)
	if (! null item) {
		Module m = module(item)
		return m
	}
	
	return null
}

bool isOpenModule(string moduleFullName) {
	return getOpenModule(moduleFullName) != null
}

bool internalCloseVerified(Module m) {
	if (! null openedModules) {
		delete(openedModules, fullName(m) "")
	}
	close(m)
	return true
}

bool internalClose(Module m) {
	if (isVisible(m)) {
		if (! unsaved(m)) {
			return internalCloseVerified(m)
		}
		else {
			return false
		}
	}
	else {
		return internalCloseVerified(m)
	}
}

bool internalSaveAndClose(string moduleFullName) {
	Module m = getOpenModule(moduleFullName);
	if (! null m) {
		if (isEdit(m) && ! isVisible(m) && unsaved(m))
			save(m)
		return internalClose(m)
	}
	else
		return false;
}

Module internalOpenModuleByName(string moduleFullName, int editMode) {
	Module m = null
	
	// already opened?
	m = getOpenModule(moduleFullName)
	if (! null m) {
		if (editMode == MODULE_MODE_EDIT) {
			if (isEdit(m))
				return m
		}
		else if (editMode == MODULE_MODE_SHARE) {
			if (isEdit(m) || isShare(m))
				return m
		}
		else
			return m
	}
	m = null
	
	if (null openedModules) {
		openedModules = createString
	}
	
	// edit?
	if (editMode == MODULE_MODE_EDIT) {
		noError(); // avoid to throw an error if the module itself contains invalid DXL code
		m = edit(moduleFullName, false, true);
		lastError();
		
		if (! null m && ! isEdit(m)) {
			if (internalClose(m)) {
				noError(); // avoid to throw an error if the module itself contains invalid DXL code
				m = edit(moduleFullName, false, true);
				lastError();
			}
		}
	}
	
	// share?
	else if (editMode == MODULE_MODE_SHARE) {
		noError(); // avoid to throw an error if the module itself contains invalid DXL code
		m = share(moduleFullName, false, true);
		lastError();
		
		if (! null m && ! isShare(m) && ! isEdit(m)) {
			if (internalClose(m)) {
				noError(); // avoid to throw an error if the module itself contains invalid DXL code
				m = share(moduleFullName, false, true);
				lastError();
			}
		}
	}
	
	// read
	else {
		noError(); // avoid to throw an error if the module itself contains invalid DXL code
		m = read(moduleFullName, false);
		lastError();
	}
		
	// mark as open
	if (! null m) {
		if (! isVisible(m)) {
			put(openedModules, moduleFullName "", moduleFullName)
		}
	}
	
	// check the module is now open
	if (null m) {
       if (editMode == MODULE_MODE_EDIT) {
           sodius_setError("Cannot open the module in exclusive edit mode: " moduleFullName);
       }
       else if (editMode == MODULE_MODE_SHARE) {
           sodius_setError("Cannot open the module in edit mode: " moduleFullName);
       }
       else {
           sodius_setError("Cannot open the module in read-only mode: " moduleFullName);
       }
	}
	
	return m
}

Module internalOpenModuleByUrl(string url, int editMode) {
	string dbHost
	int dbPort
	string dbName
	string dbId
	Item i
	ModuleVersion modVer
	int objectAbsno
	
	string result = decodeURL(url, dbHost, dbPort, dbName, dbId, i, modVer, objectAbsno) 
	if (null result) {
	    if (null modVer) {
	        sodius_setError("Invalid URL for a Module or an Object: " url);
            return null
	    }
		else if (editMode == MODULE_MODE_READ && isBaseline(modVer)) {
			noError(); // avoid to throw an error if the module itself contains invalid DXL code
			Module m = load(modVer, false);
			lastError();
			return m;
		}
		else {
			return internalOpenModuleByName(fullName(modVer), editMode)
		}
	}
	else {
	    sodius_setError("Cannot open the module " url " : " result);
		return null
	}
}

Module readModuleByName(string moduleFullName) {
	return internalOpenModuleByName(moduleFullName, MODULE_MODE_READ)
}

Module readModuleByUrl(string url) {
	return internalOpenModuleByUrl(url, MODULE_MODE_READ)
}

Module editModuleByName(string moduleFullName) {
	return internalOpenModuleByName(moduleFullName, MODULE_MODE_EDIT)
}

Module editModuleByUrl(string url) {
	return internalOpenModuleByUrl(url, MODULE_MODE_EDIT)
}

Object internalOpenObject(string url, int editMode) {
	string dbHost
	int dbPort
	string dbName
	string dbId
	Item i
	ModuleVersion modVer
	int objectAbsno
	
	string result = decodeURL(url, dbHost, dbPort, dbName, dbId, i, modVer, objectAbsno) 
	if (null result) {
        if (null modVer) {
            sodius_setError("Invalid URL for an Object: " url);
            return null
        }
        else {
		    Module m = internalOpenModuleByName(fullName(modVer), editMode)
    		if (! null m) {
    			if (objectAbsno >= 0) {
    				Object o = object(objectAbsno, m)
    				if (null o) {
    				    sodius_setError("The Object does no longer exist: " url);
    				    return null;
    				}
    				else {
    				    return o;
    				}
    			}
    			else {
                    sodius_setError("Invalid URL for an Object: " url);
                    return null
                }
    		}
		}
	}
	else {
        sodius_setError("Cannot open the Object " url " : " result);
        return null
    }
	
	return null
}

Object readObjectByUrl(string url) {
	return internalOpenObject(url, MODULE_MODE_READ)
}

Object editObjectByUrl(string url) {
	return internalOpenObject(url, MODULE_MODE_EDIT)
}

Object internalOpenObject(string moduleFullName, int objectAbsno, int editMode) {
	Module m = internalOpenModuleByName(moduleFullName, editMode)
	if (null m) {
	    return null;
	}
	else {
		Object o = object(objectAbsno, m)
		if (null o) {
            sodius_setError("The Object " objectAbsno " does no longer exist in: " moduleFullName);
            return null;
        }
        else {
            return o;
        }
	}
}

Object readObjectById(string moduleFullName, int objectAbsno) {
	return internalOpenObject(moduleFullName, objectAbsno, MODULE_MODE_READ)
}

Object editObjectById(string moduleFullName, int objectAbsno) {
	return internalOpenObject(moduleFullName, objectAbsno, MODULE_MODE_EDIT)
}

Object lockObjectById(string moduleFullName, int objectAbsno) {
	
	// open share
	Module m = internalOpenModuleByName(moduleFullName, MODULE_MODE_SHARE)
	if (! null m) {
		Object o = object(objectAbsno, m)
		if (null o) {
		    sodius_setError("The Object " objectAbsno " does no longer exist in: " moduleFullName);
		    return null;
		}
		else {
			
			if (! isShare(m))
				return o;
			
			// lock object
			string result = lock o
			if (null result) {
				if (lockedObjects == null)
					lockedObjects = create
				put(lockedObjects, o, o)
				return o
			}
			else {
				// edit
				return editObjectById(moduleFullName, objectAbsno)
			}
		}
	}
	
	return null;
}

void internalCloseModules() {
	if (! null openedModules) {
		
		// really close?
		if (openModulesFilename == null) {
			string moduleFullName
			for moduleFullName in openedModules do {
				Module m = getOpenModule(moduleFullName)
				if (! null m)
					internalClose(m)
			}
		}
		else {
			// just save open module names in a file
			Stream openModulesFile = write(openModulesFilename, MODULE_EDIT_DOORS_CODEPAGE)
			string moduleFullName
			for moduleFullName in openedModules do {
				openModulesFile << moduleFullName "\n"
			}
			close(openModulesFile)
		}
		
		delete(openedModules)
		openedModules = null
	}
}

void discardAndCloseModules() {
	
	// discard locked sections
	if (! null lockedObjects) {
		Object o
		for o in lockedObjects do {
			unlockDiscardSection(o)
		}

		delete(lockedObjects)
		lockedObjects = null;
	}
	
	// close modules
	internalCloseModules()
	
}

void saveAndCloseModules() {
	flushDeletions()
	
	// save modules
	if (! null openedModules && openModulesFilename == null) {
		string moduleFullName
		for moduleFullName in openedModules do {
			Module m = getOpenModule(moduleFullName)
			if (! null m && isEdit(m) && ! isVisible(m)) {
				if (unsaved(m))
					save(m)
			}
		}
	}
	
	// save locked sections
	if (! null lockedObjects) {
		Object o
		for o in lockedObjects do {
			unlockSaveSection(o)
		}

		delete(lockedObjects)
		lockedObjects = null;
	}
	
	// close modules
	internalCloseModules()
}
