/*
 * @param xmlFilename - output XML filename
 */
 
Stream xmlFile
Buffer xmlBuffer = create

const string EVENT_START_TAG = "0|"
const string EVENT_START_TAG_END = "1|"
const string EVENT_END_TAG = "2|"
const string EVENT_END_EMPTY_TAG = "3|"
const string EVENT_ATTRIBUTE_STRING = "4|"
const string EVENT_ATTRIBUTE_NON_STRING = "5|"

const string SEPARATOR = "|"

const string START_TAG_XMI = EVENT_START_TAG "100|"
const string END_TAG_XMI = EVENT_END_TAG "100|"

const string ATTR_XMI_ID = EVENT_ATTRIBUTE_NON_STRING "100|"
const string ATTR_XMI_VERSION = EVENT_ATTRIBUTE_STRING "101|"
const string ATTR_XMLNS_XMI = EVENT_ATTRIBUTE_STRING "102|"
const string ATTR_XMLNS_DOORS = EVENT_ATTRIBUTE_STRING "103|"
const string ATTR_XMLNS_XSI = EVENT_ATTRIBUTE_STRING "104|"

void flushXMLBuffer() {
	if (length(xmlBuffer) == 0)
		return;
		
	xmlFile << stringOf xmlBuffer
	setempty xmlBuffer;
}

void requestFlushXMLBuffer() {
	if (length(xmlBuffer) > 512) {
		flushXMLBuffer()
	}
}

void writeXMLStartTag(string tagId) {
	xmlBuffer += tagId
}

void writeXMLStartTagEnd() {
	xmlBuffer += EVENT_START_TAG_END
	requestFlushXMLBuffer();
}

void writeXMLEndTag(string tagId) {
	xmlBuffer += tagId
	requestFlushXMLBuffer();
}

void writeXMLEndTagEmpty() {
	xmlBuffer += EVENT_END_EMPTY_TAG
	requestFlushXMLBuffer();
}

void writeXMLAttributeString(string attrId, string value) {
	if (value == null || value == "")
		return;
		
	xmlBuffer += attrId
	int l = length(value)
	xmlBuffer += l SEPARATOR 
	xmlBuffer += value
	xmlBuffer += SEPARATOR
	flushXMLBuffer();
}

void writeXMLAttributeInt(string attrId, int value) {
	xmlBuffer += attrId 
	xmlBuffer += value SEPARATOR;
}

void writeXMLStartDocument() {
	xmlFile = write(xmlFilename, DOORS_CODEPAGE)
}

void writeXMLEndDocument() {
	flushXMLBuffer();
	close(xmlFile)
	delete xmlBuffer
}

void writeXMIStart() {
	writeXMLStartDocument();
	writeXMLStartTag(START_TAG_XMI);
	writeXMLAttributeString(ATTR_XMI_VERSION, "2.1")
	writeXMLAttributeString(ATTR_XMLNS_XMI, "http://www.omg.org/XMI")
	writeXMLAttributeString(ATTR_XMLNS_DOORS, "http://www.mdworkbench.com/doors")
	writeXMLAttributeString(ATTR_XMLNS_XSI, "http://www.w3.org/2001/XMLSchema-instance")
	writeXMLStartTagEnd()
}

void writeXMIEnd() {
	writeXMLEndTag(END_TAG_XMI)
	writeXMLEndDocument();
}
