/*
 * @param xmlFilename - output XML filname
 */
 
bool FORMAT_XML = false
Stream xmlFile
Buffer xmlBuffer = create
int xmlTab = 0;

void flushXMLBuffer() {
	if (length(xmlBuffer) == 0)
		return;
		
	xmlFile << stringOf xmlBuffer
	setempty xmlBuffer;
}

void requestFlushXMLBuffer() {
	if (length(xmlBuffer) > 512) {
		flushXMLBuffer()
	}
}

void cleanString(Buffer buffer, string s)
{
	int length = length(s);
	int i;
	for(i = 0; i < length; i++) {
		char c = s[i]
		if(c == '&') buffer += "&amp;"
		else if(c == '\'') buffer += "&apos;"
		else if(c == '\"') buffer += "&quot;"
		else if(c == '<') buffer += "&lt;"
		else if(c == '>') buffer += "&gt;"
		else if(c == '\n') buffer += "&#10;"
		else if(c == '\r') buffer += "&#13;"
		else if(c == '\t') buffer += "&#9;"
		else buffer += c
	}
}

void writeTabs() {
	if (FORMAT_XML) {
		int i;
		for (i = 0; i < xmlTab; i++)
			xmlBuffer += "  ";
	}
}

void writeXMLStartTag(string name) {
	writeTabs()
	xmlBuffer += '<';
	xmlBuffer += name;
	xmlTab = xmlTab + 1;
}

void writeXMLStartTagEnd() {
	xmlBuffer += ">\n";
	requestFlushXMLBuffer();
}

void writeXMLEndTag(string name) {
	xmlTab = xmlTab - 1;
	writeTabs()
	xmlBuffer += "</";
	xmlBuffer += name;
	xmlBuffer += ">\n";
	requestFlushXMLBuffer();
}

void writeXMLEndTagEmpty() {
	xmlTab = xmlTab - 1;
	xmlBuffer += "/>\n";
	requestFlushXMLBuffer();
}

void writeXMLAttributeString(string name, string value) {
	if (value == null || value == "")
		return;
		
	xmlBuffer += ' ';
	xmlBuffer += name;
	xmlBuffer += "=\"";
	cleanString(xmlBuffer, value);
	xmlBuffer += "\"";
	requestFlushXMLBuffer()
}

void writeXMLAttributeEncodedString(string name, string value) {
	xmlBuffer += ' ';
	xmlBuffer += name;
	xmlBuffer += "=\"";
	xmlBuffer += value;
	xmlBuffer += "\"";
	requestFlushXMLBuffer()
}

void writeXMLAttributeBool(string name, bool value) {
	xmlBuffer += ' ';
	xmlBuffer += name;
	xmlBuffer += "=\"";
	xmlBuffer += value "";
	xmlBuffer += "\"";
	requestFlushXMLBuffer()
}

void writeXMLAttributeInt(string name, int value) {
	xmlBuffer += ' ';
	xmlBuffer += name;
	xmlBuffer += "=\"";
	xmlBuffer += value "";
	xmlBuffer += "\"";
	requestFlushXMLBuffer()
}

string toXHTML(string value) {
	if (! isRichText(value))
		return "";
		
	Buffer buffer = create
    RichTextParagraph rp
    RichText rt
    
    bool isFormatted = false
    bool isBold = false;
    bool isItalic = false;
    bool isUnderline = false;
    bool isStrike = false;
    bool isBulletParagraph = false;

	buffer += "<html><body>";
	
	// for each paragraph
    for rp in value do {
    	
    	// bullet paragraph
    	if (rp.isBullet) {
    		if (! isBulletParagraph) {
	    		buffer += "<ul>"
    			isBulletParagraph = true;
    			isFormatted = true;
    		}
    		buffer += "<li>"
    	}
    	else {
    		if (isBulletParagraph) {
	    		buffer += "</ul>"
    			isBulletParagraph = false;
    		}
    		buffer += "<p>"
    	}
    	
    	// for each chunk
        for rt in rp do
        {
	    	// close formatting tags
	    	if (isBold && ! rt.bold) {
	    		buffer += "</b>"
	    		isBold = false;
	    	}
	    	if (isItalic && ! rt.italic) {
	    		buffer += "</i>"
	    		isItalic = false;
	    	}
	    	if (isUnderline && ! rt.underline) {
	    		buffer += "</u>"
	    		isUnderline = false;
	    	}
	    	if (isStrike && ! rt.strikethru) {
	    		buffer += "</del>"
	    		isStrike = false;
	    	}
	    		
	    	// open new formatting tags (reverse order of close)
        	if (! isStrike && rt.strikethru) {
        		buffer += "<del>"
        		isStrike = true;
        		isFormatted = true;
        	}
        	if (! isUnderline && rt.underline) {
        		buffer += "<u>"
        		isUnderline = true;
        		isFormatted = true;
        	}
        	if (! isItalic && rt.italic) {
        		buffer += "<i>"
        		isItalic = true;
        		isFormatted = true;
        	}
        	if (! isBold && rt.bold) {
        		buffer += "<b>"
        		isBold = true;
        		isFormatted = true;
        	}
        	
        	// text
        	cleanString(buffer, rt.text)
        }
        
    	// close formatting tags
    	if (isBold) {
    		isBold = false;
    		buffer += "</b>"
    	}
    	if (isItalic) {
    		isItalic = false;
    		buffer += "</i>"
    	}
    	if (isUnderline) {
    		isUnderline = false;
    		buffer += "</u>"
    	}
    	if (isStrike) {
    		isStrike = false;
    		buffer += "</del>"
    	}
    		
        // end paragraph
        if (rp.isBullet) {
    		buffer += "</li>"
        }
        else {
    		buffer += "</p>"
        }
    }
    
    if (isFormatted) {
    	
		// close bullet
		if (isBulletParagraph)
    		buffer += "</ul>"
    		
		buffer += "</body></html>";
    	return stringOf(buffer)
    }
    else {
    	return "";
    }
}

void writeXMLAttributeRichText(string name, string value, string richTextValue) {
	xmlBuffer += ' ';
	xmlBuffer += name;
	xmlBuffer += "=\"";
	
	string xhtml = toXHTML(richTextValue)
	if (xhtml != null && xhtml != "") {
		cleanString(xmlBuffer, xhtml)
	}
	else {
		cleanString(xmlBuffer, value);
	}
	xmlBuffer += "\"";
	requestFlushXMLBuffer()
}

void writeXMLAttributeRichText(string name, Object o, string attributeName) {
	string value = o.attributeName ""
	if (value == null || value == "")
		return;
		
	writeXMLAttributeRichText(name, value, richText(o.attributeName));
	requestFlushXMLBuffer()
}

void writeXMLStartDocument() {
	xmlFile = write(xmlFilename, DOORS_CODEPAGE)
	xmlBuffer += "<?xml version=\"1.0\" encoding=\""
	xmlBuffer += XML_ENCODING
	xmlBuffer += "\"?>\n"
}

void writeXMLEndDocument() {
	flushXMLBuffer()
	close(xmlFile)
	delete xmlBuffer
}

